function [] = crystalEqFitPaper()

% This function plots the crystal relaxation parameters of
% Figs. 4 and 8. The input data is taken from stretched
% exponential fits to the mean squared displacement (MSD). 
% This data is omitted for reasons of space, but can be
% generated via the attached scripts. The fit itself follows
% the form MSD(t) = a*(1-exp(-abs(t/b)^c))

% The lowest density for d=9 was never run, but it appears
% in each of the saved files with the placeholder NaN.

% Set relative base directory
currentFile = mfilename( 'fullpath' );
[pathstr,~,~] = fileparts( currentFile );
basedir = [pathstr,filesep,'..',filesep,'data',filesep,'figure4and8',filesep];

cryRho = [NaN, NaN, 2^(3/2-1),2^(4/2-1),2^(5/2-1),pi^3/(48 * sqrt(3)) / hyperBallVolume(6) * 2^6,8,16,16,2^(9/2),128/5];
rhoList = {'0.9','0.905','0.91','0.915','0.92','0.925','0.93','0.935','0.94','0.945','0.95','0.955','0.96','0.965','0.97','0.975','0.98','0.985','0.99'};

gamma = NaN(numel(rhoList),8);
tau = NaN(numel(rhoList),8);
delta = NaN(numel(rhoList),8);
gammaStd = NaN(numel(rhoList),8);
tauStd = NaN(numel(rhoList),8);
deltaStd = NaN(numel(rhoList),8);
delta0 = NaN(1,8);
delta0Err = NaN(1,8);
rho = zeros(numel(rhoList),1);

for i = 1:numel(rhoList)
    rho(i) = str2double(rhoList{i});
end

dimList = 3:10;
nP = [4000 5000 16384 23328 17496 6561 39366 81920]; 


gammaLoad = load([basedir,filesep,'gamma.dat']);
gammaStdLoad = load([basedir,filesep,'gammaStd.dat']);
deltaLoad = load([basedir,filesep,'delta.dat']);
deltaStdLoad = load([basedir,filesep,'deltaStd.dat']);
tauLoad = load([basedir,filesep,'tau.dat']);
tauStdLoad = load([basedir,filesep,'tauStd.dat']);
    
for i = 1:numel(dimList)
    gamma(:,i) = gammaLoad(:,i);
    gammaStd(:,i) = gammaStdLoad(:,i);
    delta(:,i) = deltaLoad(:,i);
    deltaStd(:,i) = deltaStdLoad(:,i);
    tau(:,i) = tauLoad(:,i);
    tauStd(:,i) = tauStdLoad(:,i);
end

figure(1)
clf; hold all
figure(2)
clf; hold all
figure(3)
clf; hold all

dimColor = [1 0 0; 6/7 0 1/7; 5/7 0 2/7; 4/7 0 3/7; 3/7 0 4/7; 2/7 0 5/7; 1/7 0 6/7; 0 0 1];
r = zeros(numel(rhoList),numel(dimList));

for i = 1:numel(dimList)
    r(:,i) = (rho.^dimList(i)*cryRho(dimList(i))/nP(i)).^(2/dimList(i));
end

delta = delta./(2*r);
deltaStd = deltaStd./(2*r);

figure(1); clf; hold all
figure(2); clf; hold all
figure(3); clf; hold all

for i = 1:numel(dimList)
    figure(1)
    try
        ind = ~isnan(tau(:,i));
        errorbar(1-rho(ind),tau(ind,i),tauStd(ind,i),'.','MarkerSize',15,'Color',dimColor(i,:))
        [res,~] = fit(1-rho(ind),tau(ind,i),'a+(x/b)^c','Lower',[0 0 1],'StartPoint',[2 2 2]);
        x = linspace(0,max(1-rho),100);
        h = plot(x,feval(res,x),'-','Linewidth',2,'Color',dimColor(i,:)); 
        set(get(get(h,'Annotation'),'LegendInformation'),'IconDisplayStyle','off');
    catch
    end
    
    figure(2)
    numFit = 4;
    errorbar(1-rho,2^(i+2)*delta(:,i),deltaStd(:,i),'.','MarkerSize',15,'Color',dimColor(i,:)); 
    x = linspace(min(1-rho),max(1-rho),100);
    try
        ind = ~isnan(delta(:,i));
        rhoTemp = rho(ind);
        deltaFit = delta(ind,i);
        [res2,~] = fit(log10(1-rhoTemp(end-numFit:end)),log10(deltaFit(end-numFit:end)),'2*x + a','startpoint',0);
        h = plot(x,2^(i+2)*10.^res2.a*x.^2,'-','Linewidth',2,'Color',dimColor(i,:));
        set(get(get(h,'Annotation'),'LegendInformation'),'IconDisplayStyle','off');
        con = confint(res2);
        delta0(i) = res2.a;
        delta0Err(i) = con(2) - res2.a; 
    catch
    end
    
    figure(3)
    try
        errorbar(1-rho,gamma(:,i),gammaStd(:,i),'.','MarkerSize',15,'Color',dimColor(i,:));
        ind = ~isnan(gamma(:,i));
        [res,~] = fit(1-rho(ind),gamma(ind,i),'poly1');
        x = linspace(0,max(1-rho),100);
        h = plot(x,feval(res,x),'-','Linewidth',2,'Color',dimColor(i,:));
        set(get(get(h,'Annotation'),'LegendInformation'),'IconDisplayStyle','off');
    catch
    end
end

figure(1)
xlabel('$1-x^{1/d}$','interpreter','latex')
ylabel('$\tau_\beta$','interpreter','latex')
plotSpecs()
set(gca,'FontSize',20)
set(gca,'Linewidth',1.5)
legend('$d=3$','$d=4$','$d=5$','$d=6$','$d=7$','$d=8$','$d=9$','$d=10$','numcolumns',2,'location','northwest','interpreter','latex')

figure(2)
xlabel('$1-x^{1/d}$','interpreter','latex')
ylabel('$2^d\Delta$','interpreter','latex')
set(gca,'yscale','log');
set(gca,'xscale','log');
plotSpecs()
set(gca,'FontSize',20)
set(gca,'Linewidth',1.5)
legend('$d=3$','$d=4$','$d=5$','$d=6$','$d=7$','$d=8$','$d=9$','$d=10$','numcolumns',2,'location','northwest','interpreter','latex','fontsize',16)
xlim([0.01 0.1])
ylim([3e-4 1e2])
xticks([0.01 0.1])
yticks([1e-3 1e-2 1e-1 1e0 1e1 1e2])

figure(3)
xlabel('$1-x^{1/d}$','interpreter','latex')
ylabel('$\gamma$','interpreter','latex')
legend('$d=3$','$d=4$','$d=5$','$d=6$','$d=7$','$d=8$','$d=9$','$d=10$','numcolumns',2,'location','southeast','interpreter','latex')
plotSpecs()
set(gca,'FontSize',20)
set(gca,'Linewidth',1.5)
axis([0 0.1 0.4 1])

figure(4)
clf; hold all
errorbar(1./(3:10)',10.^(delta0),log(10)*delta0Err,'.k','markersize',35,'linewidth',3)
xlabel('$1/d$','interpreter','latex')
ylabel('$\Delta_0$','interpreter','latex')
plotSpecs()
set(gca,'FontSize',20)
set(gca,'Linewidth',1.5)
axis([0 0.35 0.5 1.5])

end