import monteCarloPCP as mc
import numpy as np
from scipy.special import gamma
import Resources.crystalStructures.e8Crystal as e8
import Resources.crystalStructures.e7Crystal as e7
import Resources.crystalStructures.e6Crystal as e6
import Resources.crystalStructures.dDCrystal as dD
import Resources.crystalStructures.p10cCrystal as p10c
import sys
import os
import glob
import time

# This function creates a series of 'numAvgs' crystal configurations saved and pre-averaged at a 
# frequency set by the variable 'saveFreq'. This function ultimately saves the pair correlation 
# function g(r) as gofr.dat. This g(r) is log spaced, has 'numR' entries, and a maximum distance
# value at approximately the first minimum. It is created at dimension 'dim', with number of cells 
# 'numCells', which can be translated to N below, and number density (N/V) 'rhoString'. While in
# principle the neighbors need not be recalculated int eh solid, checkNeighbors=1 sets the standard
# distance argument to calculate neighbors once one has broken out of its neighbor range.
# basedir specifies where it will be saved. In accordance with the data here, it should be given 
# as ../data/figure2/crystalData/.

# It is called as 
# python3 generateCrystal.py dim numCells rhoString numAvgs saveFreq checkNeighbors basedir


dim = int(sys.argv[1])
numCells = int(sys.argv[2])
rhoString = sys.argv[3]
numAvgs = int(sys.argv[4])
saveFreq = int(sys.argv[5])
checkNeighbors = int(sys.argv[6])==1
basedir = sys.argv[7]

# E boundaries are boundaryType = 1
# D boundaries are boundaryType = 2
# Z boundaries are boundaryType = 3
# volpoints give sthe number of points used when numerically integrating g(r) when r > L/2 
# (with L = box length)
boundaryType = 3
rho = np.float64(rhoString)
if dim==6:
	# In 6d, we must tile our box to be close to orthorhombic. numCells provides a key to which
	# option below we would like to use. Check side lengths as 'sides'
	if numCells == 19:
		numCells = np.array([3,3,3,3,3,3])
		crystalPos, sides = e6.generateE6Crystal(numCells)
	elif numCells == 20:
		numCells = np.array([2,2,2,2,2,2])
		crystalPos, sides = e6.generateE6Crystal(numCells)
	elif numCells == 21:
		numCells = np.array([2,2,2,2,2,3])
		crystalPos, sides = e6.generateE6Crystal(numCells)
	elif numCells == 22:
		numCells = np.array([2,2,2,2,3,3])
		crystalPos, sides = e6.generateE6Crystal(numCells)
	elif numCells == 23:
		numCells = np.array([2,2,2,3,3,3])
		crystalPos, sides = e6.generateE6Crystal(numCells)
	elif numCells == 24:
		crystalPos, sides = e6.generateE6Crystal(numCells,version=2)
	else:
		crystalPos, sides = e6.generateE6Crystal(numCells,version=3)
	maxPhi = e6.maxPhi()
	volPoints = 1000000
elif dim==7:
	# E7 has only one valid configuration encoded, all numCells yield the same result
	crystalPos, sides = e7.generateE7Crystal()
	maxPhi = e7.maxPhi()
	volPoints = 1000000
elif dim==8:
	# version = 1 gives E8 boundaries with N=6561
	# version = 2 gives D8 boundaries with N=13122
	# version = 3 gives Z8 boundaries with N=65536 
	crystalPos, sides = e8.generateE8Crystal(version = numCells)
	maxPhi = e8.maxPhi()
	volPoints = 1000000
	# boundary type corresponds with version number
	boundaryType = numCells
elif dim==10:
	# version = 1 gives D10^+ boundaries with N=40960
	# version = 2 gives D10 boundaries with N=81920
	crystalPos, sides = p10c.generateP10cCrystal(version = numCells)
	maxPhi = p10c.maxPhi()
	volPoints = 1000000
	# boundary type corresponds with version number
	boundaryType = numCells
else:
	# numCells > 100 gives Dn boundaries with N=(numCells - 100)^d
	# numCells < 100 gives Zn boundaries with N=0.5*(2*numCells)^d
	if numCells < 100:
		crystalPos, sides = dD.generatedDCrystal(numCells,dim)
		maxPhi = dD.maxPhi(dim)
	else:
		boundaryType = 2
		crystalPos, sides = dD.generatedDCrystalDBC(numCells-100,dim)
		maxPhi = dD.maxPhi(dim)
		boxVolume = 2
	volPoints = 1000000	

nParticles = crystalPos.shape[0]

saveDir = basedir + '/' + str(dim) + '/' + rhoString + '/' + str(nParticles) + '/'

os.makedirs(saveDir,exist_ok=True)


# Initialize packing
p = mc.MCPacking(dim,nParticles,1.0)
phi = rho*p.hyperBallCoefficient(dim)/(2**dim)
p.setPhi(phi)

if boundaryType == 3:
	boxVolume = 1
elif boundaryType==2:
	boxVolume = 2
	p.setGeometryType(mc.geometryEnum.Dn)
else: #boundaryType==1
	boxVolume = 1
	p.setGeometryType(mc.geometryEnum.E8)

p.setPhi(phi)

maxRad = (maxPhi * boxVolume/(nParticles * p.hyperBallCoefficient(dim)))**(1/dim)
p.setBoxSize(sides)
#maxDist = 4*(maxRad - p.getRadii()[0]) + 2*p.getRadii()[0]
maxDist = ((maxRad/p.getRadii()[0] - 1) * 4 + 1) * 2 * p.getRadii()[0]
cutDist = maxDist/(2*p.getRadii()[0]) - 1
print(p.getRadii()[0])

logBinMin = np.log10(1/(nParticles))

if dim == 8:
	cutDist = 1
elif dim >= 9:
	cutDist *= 1
elif dim == 7:
	cutDist *= 1.5
elif dim == 6:
	cutDist *= 2
else:
	cutDist *= 3

print(cutDist)
p.setPositions(crystalPos)
if dim < 9:
	p.calcNeighborsCut(cutDist)

if dim ==10:
	p.calcNeighborsCut(cutDist)

minRadius = np.min(p.getRadii())

# Find the maximum MC displacement or load it if it exists
if os.path.exists(saveDir + '/maxMCDisp.dat'):
	with open(saveDir + '/maxMCDisp.dat', 'r') as file:
		maxDisp = np.float(file.read())
else:
	if dim == 9:
		# maxDisp = (maxRad - p.getRadii()[0])/3
		maxDisp = p.findMCMaxDisplacement(hiDist = maxDist, distTol = 1e-4, neighborCut = cutDist,verbose = True,numTrials=10)
		p.setMCMaxDisplacementD9Relative(3*maxDisp)
		p.setMCMaxDisplacementCM(maxDisp)
	else:
		maxDisp = p.findMCMaxDisplacement(hiDist = maxDist, distTol = 1e-4, neighborCut = cutDist,verbose = True,numTrials=40)
	with open(saveDir + '/maxMCDisp.dat', 'w') as file:
		file.write(str(maxDisp) + '\n')



p.setMCMaxDisplacement(maxDisp)
p.resetCenterOfMass()
p.setPositions(crystalPos)
numR = 75
g = np.zeros(numR)
r = np.zeros(numR)
# This is set to be larger than 3*tau_alpha. Should adjust based on dimension and phi by checking
# the average time for the MSD to plateau.
numMSD = 400
p.setRandomSeed(np.random.randint(2**16))
# Start where the simulation left off
start = len(glob.glob(saveDir + '/*/'))
print("calculating neighborlist")
p.calcNeighborsCut(cutDist);
p.calcForceEnergy()
p.resetDisplacement()
saveNum = 0
i = 1
neigh = p.getNeighbors()
print("neighbors: " + str(neigh.nnz/(dim*nParticles)))


print(saveDir + '/' + str(saveNum) + '/gofr.dat')
while saveNum < numAvgs:
	msd = np.zeros(numMSD)
	startingPos = p.getUnwrappedPositions()
	p.resetCenterOfMass()
	for j in range(numMSD):
		#Should check how far particles have moved and then recalculate neighbors if it's beyond max mc disp
		if (checkNeighbors and p.getSumOfMaxDisplacements() >= 2*minRadius*cutDist):
			print('resetting at ' + str(j))
			p.calcNeighborsCut(cutDist)
			p.resetDisplacement()
		p.MCStep()
		delta = p.getUnwrappedPositions() - p.getCenterOfMassShift() - startingPos
		msd[j] = np.mean(np.sum((delta - np.round(delta))**2,1))
		msd[j] = np.mean(np.sum(delta**2,1))
        mc.fileIO.saveArray(saveDir + '/' + str(i) + '/msd.dat',msd)
	# p.calcNeighborsCut(cutDist)
	thisg,thisr = p.gofr(numR,maxRad = maxDist,minRad = p.getRadii()[0]*2,numVolumePoints = volPoints,logBin = True,logBinMin = logBinMin)
	thisg[np.isnan(thisg)] = 0
	thisg[np.isinf(thisg)] = 0
	r = thisr/(2*p.getRadii()[0])
	g = g+thisg
	if (i%saveFreq == 0):
		saveNum = len(glob.glob(saveDir + '/*/'))
		os.makedirs(saveDir + '/' + str(saveNum),exist_ok=True)
		mc.fileIO.save2DArray(saveDir + '/' + str(saveNum) + '/gofr.dat',np.stack((thisr/(2*p.getRadii()[0]),g/saveFreq),axis=1))
		# Save MSD if it is desired by uncommenting the below line:
		# mc.fileIO.saveArray(saveDir + '/' + str(saveNum) + '/msd.dat',msd)
		g = np.zeros(numR)
	i = i + 1

