import monteCarloPCP as mc
import numpy as np
import sys
import os
import glob
import time
import Resources.crystalStructures.e8Crystal as e8
import Resources.crystalStructures.e7Crystal as e7
import Resources.crystalStructures.e6Crystal as e6
import Resources.crystalStructures.dDCrystal as dD
import Resources.crystalStructures.p10cCrystal as p10c

# This function creates a series of 'numAvgs' liquid configurations saved and pre-averaged at a 
# frequency set by the variable 'saveFreq'. This function ultimately saves the pair correlation 
# function g(r) as gofr.dat. This g(r) is linearly spaced, has 'numR' entries, and a maximum distance
# value at approximately the second peak height if 'fullDist' = 0 and at the maximum possible distance
# if 'fullDist' = 1. It is created at dimension 'dim', number of particles 'nParticles', and number 
# density (N/V) 'rhoString'. boundaryType is 1 for D boundary condtions, 2 for E, and anything else for
# Z boundary conditions. basedir specifies where it will be saved. In accordance with the data here, it
# should be given as ../data/figure2/liquidData/.

# It is called as 
# python3 generateLiquid.py dim nParticles rhoString numR saveFreq numAvgs fullDist boundaryType basedir



dim = int(sys.argv[1])
nParticles = int(sys.argv[2])
rhoString = sys.argv[3]
numR = int(sys.argv[4])
saveFreq = int(sys.argv[5])
numAvgs = int(sys.argv[6])
fullDist = int(sys.argv[7])
boundaryType = int(sys.argv[8])
basedir = sys.argv[9]


if dim==6:
	maxPhi = e6.maxPhi()
elif dim==8:
	maxPhi = dD.maxPhi(dim)
elif dim==9:
	maxPhi = dD.maxPhi(dim)
elif dim==10:
	maxPhi = p10c.maxPhi()
elif dim == 7:
	maxPhi = e7.maxPhi()
else:
	maxPhi = dD.maxPhi(dim)

rho = np.float64(rhoString)
volPoints = 1000000
saveDir = basedir + '/' + str(dim) + '/' + rhoString + '/' + str(nParticles) + '/'
tempDir = saveDir + '/temp/'

# initialize structure
p = mc.MCPacking(dim,nParticles,0.1)

# It is somewhat difficult to equilibrate monodisperse 3d spheres near coexistence while avoiding
# crystalization. For this reason, we use the liquid equation of state provided and do not rely
# on our own 3d liquid monodisperse data. 
if dim == 3:
	p.setBiRadii()

if (boundaryType==1):
	boxVolume = 2
	p.setGeometryType(mc.geometryEnum.Dn)
elif(boundaryType==2):
	boxVolume = 2
	p.setGeometryType(mc.geometryEnum.E8)
else:
	boxVolume = 1


phi = rho * p.hyperBallCoefficient(dim)/(2.0**dim)
p.setPhi(phi)
maxRad = (boxVolume * maxPhi/(nParticles * p.hyperBallCoefficient(dim)))**(1/dim)

# Approximate furthest distance particles should be away from one another in a crystal
maxDist = 4*(maxRad - p.getRadii()[0]) + 2*p.getRadii()[0]

# Alternate estimate of maxDist
# maxDist = ((maxRad/p.getRadii()[0] - 1) * 4 + 1) * 2 * p.getRadii()[0]
cutDist = 1

# Begin with Poisson distributed points, inflate to become soft spheres, andinimize the energy 
# to make them hard spheres
p.setRandomPositions()
p.calcNeighborsCut(cutDist)
p.calcForceEnergy()
p.minimizeFIRE(criticalForce = 1e-12)
p.calcForceEnergy()

print("volume is " + str(p.getVolume()))

minRadius = np.min(p.getRadii())

os.makedirs(saveDir,exist_ok=True)

# Load old run if it already exists
if os.path.exists(saveDir + '/tauAlpha.dat'):
	with open(saveDir + '/neighborCutDistance.dat', 'r') as file:
		neighborCutDistance = np.float(file.read())
	with open(saveDir + '/maxMCDisp.dat', 'r') as file:
		maxMCDisp = np.float(file.read())
	with open(saveDir + '/tauAlpha.dat', 'r') as file:
		tauAlpha = int(file.read())
	p.load(tempDir)
else:
	# thermalize packing to 3*tau_alpha steps
	numSteps, tauAlpha, neighborCutDistance, maxMCDisp, msd, _ = p.thermalizePacking(verbose=True,numTrials=10,numCosts=5,numTau=3)
	p.save(tempDir)
	with open(saveDir + '/maxMCDisp.dat', 'w') as file:
		file.write(str(numSteps) + '\n')
	with open(saveDir + '/neighborCutDistance.dat', 'w') as file:
		file.write(str(neighborCutDistance) + '\n')
	with open(saveDir + '/tauAlpha.dat', 'w') as file:
		file.write(str(tauAlpha) + '\n')

g = np.zeros(numR)
r = np.zeros(numR)
p.setRandomSeed(np.random.randint(2**16))

# set iterator to the last generated saved position
start = len(glob.glob(saveDir + '/[0-9]*/'))
p.calcNeighborsCut(neighborCutDistance);
p.calcForceEnergy()
p.resetDisplacement()
saveNum = start
for i in range(int(saveFreq*saveNum)+1,numAvgs+1):
	p.calcNeighborsCut(neighborCutDistance)
	for j in range(3*tauAlpha):
		#Should check how far particles have moved and then recalculate neighbors if it's beyond max mc disp
		if (p.getSumOfMaxDisplacements()/minRadius >= 2*neighborCutDistance):
			print('resetting at ' + str(j))
			p.calcNeighborsCut(neighborCutDistance)
			p.resetDisplacement()
		# One MC step
		p.MCStep()
	p.calcNeighborsCut(10)
	if fullDist == 1:
		thisg,thisr = p.gofr(numR,minRad = p.getRadii()[0]*2,numVolumePoints = volPoints,logBin = False)
	else:
		thisg,thisr = p.gofr(numR,maxRad = maxDist,minRad = p.getRadii()[0]*2,numVolumePoints = volPoints,logBin = False)
	thisg[np.isnan(thisg)] = 0
	thisg[np.isinf(thisg)] = 0
	r = thisr/(2*p.getRadii()[0])
	g = g+thisg
	if (i%saveFreq == 0):
		os.makedirs(saveDir + '/' + str(saveNum),exist_ok=True)
		mc.fileIO.save2DArray(saveDir + '/' + str(saveNum) + '/gofr.dat',np.stack((thisr/(2*p.getRadii()[0]),g/saveFreq),axis=1))
		g = np.zeros(numR)
		saveNum += 1
		p.save(tempDir)

