import numpy as np
import warnings
from numpy import linalg
from scipy import sparse, spatial

from .. import enums

class Mixin():
    def getStableListAndExcess(self,checkConvexHull=None,gap=0.0):
        """
        Calculate the list of stable particles and the number of excess
        contacts beyond isostaticity

        args:
            gap (numpy quad): The gap cutoff used to determine whether
            particles are in contact. Defaults to zero.
        returns:
            stableList (numpy.bool array): boolean mask equal to True for
                stable particle indices. For a list of indices, use
                np.where(stableList)
            excess: number of excess contacts, excluding rattlers.
        """
        # Contact vectors are stored as bsr for efficient row access
        if checkConvexHull is None:
            if self.getNDim() > 5:
                # because convex hull takes exponential time in dimension, it begins to take forever
                # forever beginning around 10 dimensions. also, it's arguably not useful to check
                # for co-hemispheric contacts anyway because they never actually happen.
                warnings.warn("Default behavior is to not check for co-hemispheric contacts in d > 5")
                checkConvexHull = False
            else:
                checkConvexHull = True
        cv = self.getContactVectors(gap=gap).astype(np.float64).tobsr(
            (1, self.getNDim())
            )
        # All particles are initiially assumed to be stable
        stableList = np.ones(self.getNumParticles(), dtype=np.bool)
        maxContacts = np.max(cv.indptr[1:] - cv.indptr[:-1])
        A = np.zeros((maxContacts + 1, self.getNDim()), dtype=np.float64)
        # Loop until no updates are made
        unchanged = False
        while not unchanged:
            unchanged = True
            # Remove unstable contacts using a partial identity
            cv @= sparse.diags(
                np.repeat(stableList, self.getNDim()), dtype=np.bool
                ).tobsr((self.getNDim(),) * 2)
            for i in range(self.getNumParticles()):
                if stableList[i]:
                    # Get nonzero elements in row
                    row = cv.data[cv.indptr[i]:cv.indptr[i+1], 0, :]
                    numContacts = row.shape[0]
                    if numContacts < self.getNDim() + 1:
                        stableList[i] = False
                        unchanged = False
                    elif checkConvexHull:
                        # A contains the contact vectors and the origin
                        # If the particles is stable, the origin will be an
                        # interior point
                        A[1:numContacts+1] = row
                        hull = spatial.ConvexHull(A[:numContacts+1])
                        if 0 in hull.vertices:
                            stableList[i] = False
                            unchanged = False
        # Remove unstable rows, leaving only stable contacts
        cv = sparse.diags(stableList, dtype=np.bool, format="bsr") @ cv

        if(np.sum(stableList) > 0):
            #Nc - ((N - 1) * d + 1)
            numExcess = cv.indptr[-1] // 2 - ((np.sum(stableList) - 1) * self.getNDim() + 1)
        else:
            #otherwise it returns d-1 which is confusing
            numExcess = np.nan

        return (stableList,numExcess)
