import os
import numpy as np
from .. import fileIO

"""
packingData is a tuple that controls the read/write behavior of
Packing.SavePacking and Packing.loadPacking. Its entries are tuples in the
following form:
	 name (str): datum name, should match internal name when possible
	 fileName (str): name of file for non-scalars, "name.extension" preferred
	 isScalar (bool): True if the value is a scalar
	 isMandatory (bool): If True, the value will be saved automatically, and
		  loaded automatically if loadable
	 saver (function): function acceptint the arguments packing(the packing
		  object) and fileName (the file in which it is being saved) that
		  computes the data and writes it to an appropriately named file
		If the datum is a scalar, the functuon requires only the packing as an
		argument, and returns a string
	loader (function): function accepting the arguments packing (the packing
		object) and fileName (the file from which it is being loaded) and
		loads the data onto the packing object
		If the datum is a scalar, the function requires the packing and a
		string representing the value to be set.
		Set to None if the datum is not loadable.
File names should be the same as dictionary keys. Attributes will be loaded
in accordance with their order in packingData
savers and loaders should be defined using lambda syntax when possible, but
more complex functions can be defined outside of the tuple literal. See below
"""
packingData = (
	# Basic data (these should remain first) ----------------------------------
	(
		"nDim", None, True, True,
		lambda packing: str(packing.getNDim()),
		lambda packing, ndString: packing.setNDim(int(ndString))
	),
	(
		"numParticles", None, True, True,
		lambda packing: str(packing.getNumParticles()),
		lambda packing, npString: packing.setNumParticles(int(npString))
	),

	(
		"positions", "positions.dat", False, True,
		lambda packing, fileName: fileIO.save2DArray(
			fileName, packing.getPositions()
		),
		lambda packing, fileName: packing.setPositions(
			fileIO.load2DArray(fileName, dtype=np.float64)
		)
	),

	(
		"radii", "radii.dat", False, True,
		lambda packing, fileName: fileIO.saveArray(
			fileName, packing.getRadii()
		),
		lambda packing, fileName: packing.setRadii(
			fileIO.load2DArray(fileName, dtype=np.float64)
		)
	),
    (
        "boxSize", "boxSize.dat", False, True,
        lambda packing, fileName: fileIO.saveArray(
            fileName, packing.getBoxSize()
        ),
        lambda packing, fileName: packing.setBoxSize(
            fileIO.loadArray(fileName, dtype=np.float64)
        )
    ),
	(
		"phi", None, True, True,
		lambda packing: str(packing.getPhi()),
		None
	),
	(
		"potentialPower", None, True, True,
		lambda packing: str(packing.getPotentialPower()),
		lambda packing, ppString: packing.setPotentialPower(float(ppString))
	),
	# Basic optional data -----------------------------------------------------
	(
		"neighbors", "neighbors.mtx", False, False,
		lambda packing, fileName: fileIO.saveSparseArray(
			fileName, packing.getNeighbors()
		),
		lambda packing, fileName: packing.setNeighbors(
			fileIO.loadSparseArray(fileName, dtype=np.int64)
		)
	),
	(
		"contactVectors", "contactVectors.mtx", False, False,
		lambda packing, fileName: fileIO.saveSparseArray(
			fileName, packing.getContactVectors(normalized = True)
		),
		None
	),
	# unnormalized contact vectors:
	# this pathetic/obscure name shall remain as long as
	# the unit contact vectors are called contact vectors in packing.cu
	(
		"contactVectorsUN", "contactVectorsUN.mtx", False, False,
		lambda packing, fileName: fileIO.saveSparseArray(
			fileName, packing.getContactVectors(normalized = False)
		),
		None
	),
	(
		"contacts", "contacts.mtx", False, False,
		lambda packing, fileName: fileIO.saveSparseArray(
			fileName, packing.getContacts()
		),
		None
	),
	(
		"excessContacts", None, True, False,
		lambda packing: str(packing.getStableListAndExcess()[1]),
		None
	),
	(
		"energy", None, True, False,
		lambda packing: str(packing.getEnergy()),
		None
	),
	(
		"maxUnbalancedForce", None, True, False,
		lambda packing: str(packing.getMaxUnbalancedForce()),
		None
	),
	(
		"stableList", "stableList.dat", False, False,
		lambda packing, fileName: fileIO.saveArray(
			fileName, packing.getStableListAndExcess()[0].astype(np.int32)
		),
		None
	),
	(
		"netForces", "netForces.dat", False, False,
		lambda packing, fileName: fileIO.save2DArray(
			fileName, packing.getNetForces(recalc=True)
		),
		None
	),
	(
		"allForces", "allForces.mtx", False, False,
		lambda packing, fileName: fileIO.saveSparseArray(
			fileName, packing.getAllForces()
		),
		None
	),
	(
		"velocities", "velocities.dat", False, False,
		lambda packing, fileName: fileIO.save2DArray(
			fileName, packing._getVelocities()
		),
		lambda packing, fileName: packing._setVelocities(
			fileIO.load2DArray(fileName, dtype=np.float64)
		)
	),
	(
		"forces", "forces.dat", False, False,
		lambda packing, fileName: fileIO.save2DArray(
			fileName, packing.getForces()
		),
		lambda packing, fileName: packing.setForces(
			fileIO.load2DArray(fileName, dtype=np.float64)
		)
	),
	(
		"lastPositions", "lastPositions.dat", False, False,
		lambda packing, fileName: fileIO.save2DArray(
			fileName, packing._getLastPositions()
		),
		lambda packing, fileName: packing._setLastPositions(
			fileIO.load2DArray(fileName, dtype=np.float64)
		)
	),
	(
		"stableForceMagnitudes", "stableForceMagnitudes.mtx", False, False,
		lambda packing, fileName: fileIO.saveSparseArray(
			fileName, packing.getForceMagnitudes(stable = True)
		),
		None
	),
	(
		"overlaps", "overlaps.dat", False, False,
		lambda packing, fileName: fileIO.saveSparseArray(
			fileName, packing.getSmallOverlaps()
		),
		None
	),

)


class Mixin():
	def save(self, path, optionalData=[], overwrite=False):
		"""
		Saves the packing to a directory.  When finished write the "isFinished 1"
		data to the scalars.dat file to indicate that everything has been
		successfully saved.
		args:
			path (str): path to output directory
			optionalData (list of str): flags for optional data do be saved.
				This list is compared with packingData for an appropriate
				reader/writer.
			overwrite (bool): When True, savePacking will delete the contents
				of an existing Packing directory if it encounters one.
		"""
		# Paths ending in a separator are not compatible with this syntax
		path.rstrip(os.sep)
		# Checks that all data options are acceptable
		for od in optionalData:
			if od not in (pd[0] for pd in packingData):
				raise ValueError(
					"option {0} did not match an existing signature. "
					"Acceptable data options are:\n\t{1}"
					.format(od, "\n\t".join(pd[0] for pd in packingData))
					)
		if os.path.isdir(path):
			if overwrite:
				for fileName in os.listdir(path):
					if os.path.isfile(fileName):
						os.remove(path + os.sep + fileName)
			else:
				raise FileExistsError(
					"Directory {0} already exists. set overwrite=True to "
					"delete existing contents".format(path)
					)
		else:
			# Use mkdirs to allow for creating the whole directory structure
			# rather than just the leaf
			os.makedirs(path)
		with open(path + os.sep + "scalars.dat", "w") as scalars:
			for (
					name, fileName, isScalar, isMandatory, saver, loader
					) in packingData:
				if isMandatory or name in optionalData:
					if isScalar:
						scalars.write(name + "\t" + saver(self) + "\n")
					else:
						saver(self, path + os.sep + fileName)
		with open(path + os.sep + "scalars.dat", "a") as scalars:
			scalars.write("isFinished" + "\t" + "1" + "\n")

	def load(self, path, verbose=False):
		# Paths ending in a separator are not compatible with this syntax
		path.rstrip(os.sep)
		# scalars are loaded onto a dictionary
		scalarStrings = {}
		with open(path + os.sep + "scalars.dat") as file:
			for line in file:
				name, scalarString = line.strip().split("\t")
				scalarStrings[name] = scalarString
		for (
				name, fileName, isScalar, isMandatory, saver, loader
				) in packingData:
			if isScalar and loader is not None:
				if name in scalarStrings:
					if verbose:
						print("setting", name, "to", scalarStrings[name])
					loader(self, scalarStrings[name])
				elif isMandatory:
					print(
						"monteCarloPacking.Packing.load Warning: mandatory data "
						"{0} not present. This may result in undefined "
						"behavior".format(name)
						)
			elif loader is not None:
				if fileName in os.listdir(path):
					if verbose:
						print("setting", name)
					loader(self, path + os.sep + fileName)
				elif isMandatory:
					print(
						"monteCarloPacking.Packing.load Warning: mandatory data "
						"{0} not present. This may result in undefined "
						"behavior".format(name)
						)

	def loadPacking(self, fileName):
		"""
		Loads a packing file, including positions, radii and all other saved
		attributes onto a packing object
		args:
			fileName (str): Path to output file
			sep (str): whitespace separator, default "\t"
			[Packing variable name] (bool): When true, explicitly saves the
				variable to the file
		"""
		positionsRadii, flags = fileIO.read2DArray(fileName)
		self.setNDim(len(positionsRadii[0]) - 1)
		self.setNumParticles(len(positionsRadii))
		self.setPositions([line[:-1] for line in positionsRadii])
		self.setRadii([line[-1] for line in positionsRadii])
		for key, value in flags.items():
			if key == "boxSize":
				self.setBoxSize(value)
			elif key == "potentialPower":
				self.setPotentialPower(value)