#include "MCPacking.h"

#include "PackingImplementation.h"
#include "PIMPLClasses.h"
#include "FIRE.h"
#include "enumTypes.h"
#include "defs.h"
#include <algorithm>
#include <fstream>
#include <functional>
#include <iomanip>
#include <iostream>
#include <math.h>
#include <random>
#include <stdlib.h>
#include <string>
#include <sstream>
#include <vector>
#include <limits>
#include <tgmath.h> //For gamma function
#include <cmath>  //For E8 Lattice
#include <cstdint>  //For E8 Lattice

#include <tuple>
#include <utility>

#include <omp.h>

using namespace std;


MCPacking::MCPacking(int dim_i, long numParticles_i, double phi_i)
	: pi_(spimpl::make_impl<PackingImplementation>(this)),
	fire_(spimpl::make_impl<FIRE>(this)) {
	pi_->nDim = dim_i;
	pi_->numParticles = numParticles_i;
	contactList.resize(pi_->numParticles);
	neighborList.resize(pi_->numParticles);
	pi_->lastNeighborCut = 1.0;
	pi_->volume = 1.0;
	einsteinLambda = 0.0;
	pi_->pos.resize(pi_->nDim * pi_->numParticles, 0.0);
	pi_->radii.resize(pi_->numParticles, 1.0);
	setPhi(phi_i);
	pi_->potentialPower = 2.0;
	temperature = 0.0;
	initForceEnergy();
	maxDisplacement = HCMaxDispSet[pi_->nDim];
	maxDisplacementCM = 0.0;
	maxVolumeShift = 0.0;
	acceptanceRatio = 0.0;
	ecmcLength = 0.0;
	ecmcPressure = 0.0;
	d9Shift.resize(9,0.0);
	mcmcBlockedDirection.resize(pi_->nDim,false);
	pi_->lastPos.resize(pi_->nDim * pi_->numParticles,0.0);
	pi_->maxForceRescale = 0.1;
	pi_->totalDisplacement.resize(pi_->nDim * pi_->numParticles,0.0);
	pi_->distTravelledSq.resize(pi_->numParticles,0.0);
	pi_->crystalPos.resize(pi_->nDim * pi_->numParticles,0.0);
	pi_->centerOfMassShift.resize(pi_->nDim,0.0);
	pi_->centerOfMass.resize(pi_->nDim,0.0);
	pi_->unwrappedPos.resize(pi_->nDim * pi_->numParticles,0.0);
	pi_->boxSize.resize(pi_->nDim,1.0);
	pi_->constraint.resize(pi_->nDim * pi_->numParticles,0.0);
	pi_->externalForce.resize(pi_->nDim * pi_->numParticles,0.0);
	pi_->LEshift = 0.0;
	pi_->nonAdditivity = 0.0;
	pi_->overlapParameter = 0.3;
	pi_->simControl.neighborType = simControlStruct::neighborEnum::allToAll;
	pi_->simControl.distanceType = simControlStruct::distanceEnum::normal;
	pi_->simControl.geometryType = simControlStruct::geometryEnum::normal;
	pi_->simControl.potentialType = simControlStruct::potentialEnum::contactPower;
	pi_->simControl.constraintType = simControlStruct::constraintEnum::normal;
	pi_->simControl.particleType = simControlStruct::particleEnum::sphere;
	pi_->simControl.monteCarloType = simControlStruct::monteCarloEnum::metropolis;
	pi_->simControl.ensembleType = simControlStruct::ensembleEnum::nvt;
}

MCPacking::~MCPacking() = default;

long MCPacking::getNumParticles(){
	return pi_->numParticles;
}

int MCPacking::getNDim(){
	return pi_->nDim;
}

void MCPacking::setRandomSeed(long seedValue){
	mt.seed(seedValue);
}

void MCPacking::setRandomPositions(){
	switch(pi_->simControl.geometryType) {
		case simControlStruct::geometryEnum::E8:
			for (long pID = 0; pID < pi_->numParticles; pID++){
				for (int dimIndex = 0; dimIndex < pi_->nDim; dimIndex++){
					pi_->pos[pID * pi_->nDim + dimIndex] = uniformDist(mt)*2.0 - 1.0;
					pi_->unwrappedPos[pID * pi_->nDim + dimIndex] = pi_->pos[pID * pi_->nDim + dimIndex];
				}
				vector<int32_t> mirror(pi_->nDim,0);
				e8LatticePoint(pi_->pos.begin() + pID * pi_->nDim,mirror);
				for (int dimIndex = 0; dimIndex < pi_->nDim; dimIndex++){
					pi_->unwrappedPos[pID * pi_->nDim + dimIndex] = pi_->pos[pID * pi_->nDim + dimIndex];
				}
			}			
			break;
		case simControlStruct::geometryEnum::Dn:
			for (long pID = 0; pID < pi_->numParticles; pID++){
				for (int dimIndex = 0; dimIndex < pi_->nDim; dimIndex++){
					pi_->pos[pID * pi_->nDim + dimIndex] = uniformDist(mt)*2.0 - 1.0;
					pi_->unwrappedPos[pID * pi_->nDim + dimIndex] = pi_->pos[pID * pi_->nDim + dimIndex];
				}
				vector<int32_t> mirror(pi_->nDim,0);
				dnLatticePoint(pi_->pos.begin() + pID * pi_->nDim,mirror);
				for (int dimIndex = 0; dimIndex < pi_->nDim; dimIndex++){
					pi_->unwrappedPos[pID * pi_->nDim + dimIndex] = pi_->pos[pID * pi_->nDim + dimIndex];
				}
			}			
			break;
		default:
			for (long pID = 0; pID < pi_->numParticles; pID++){
				for (int dimIndex = 0; dimIndex < pi_->nDim; dimIndex++){
					pi_->pos[pID * pi_->nDim + dimIndex] = uniformDist(mt)*pi_->boxSize[dimIndex];
					pi_->unwrappedPos[pID * pi_->nDim + dimIndex] = pi_->pos[pID * pi_->nDim + dimIndex];
				}
			}
			break;
	}
}

void MCPacking::setRandomMKLambdas(){
	if(pi_->simControl.distanceType == simControlStruct::distanceEnum::MK){
		pi_->MKLambdas.resize(pi_->numParticles*pi_->numParticles*pi_->nDim);
		for (long i = 0; i < pi_->numParticles-1; i++){
			for (long j = i+1; j < pi_->numParticles; j++){
				for (int dim = 0; dim < pi_->nDim; dim++){
					pi_->MKLambdas[i*pi_->numParticles*pi_->nDim + j*pi_->nDim + dim] = uniformDist(mt)-0.5;
					pi_->MKLambdas[j*pi_->numParticles*pi_->nDim + i*pi_->nDim + dim] = -pi_->MKLambdas[i*pi_->numParticles*pi_->nDim + j*pi_->nDim + dim];
				}
			}
		}
		for (long i = 0; i < pi_->numParticles; i++){
			for (int dim = 0; dim < pi_->nDim; dim++){
				pi_->MKLambdas[i*(pi_->numParticles+1)*pi_->nDim + dim] = 0.0;
			}
		}
	}else{
		cout << "Must use MK distanceType to set lambdas" << endl;
	}
}

vector<double> MCPacking::getNetForces(){
	return pi_->netForce;
}

vector<double> MCPacking::getCenterOfMassShift(){
	return pi_->centerOfMassShift;
}

void MCPacking::initForceEnergy(){
	energyVector.resize(pi_->numParticles, 0.0);
	pi_->netForce.resize(pi_->nDim * pi_->numParticles, 0.0);
}

void MCPacking::setInitialPositions(){
	initialPositions = pi_->pos;
}

long MCPacking::getTauAlpha(){
	return tauAlpha;
}

void MCPacking::setInversePower(int inversePower_i){
	pi_->inversePower = inversePower_i;
}

int MCPacking::getInversePower(){
	return pi_->inversePower;
}

void MCPacking::setInversePowerConstants(double inversePowerCutoff_i, double inversePowerC0_i, double inversePowerC1_i, double inversePowerC2_i){
	pi_->inversePowerCutoff = inversePowerCutoff_i;
	pi_->inversePowerC0 = inversePowerC0_i;
	pi_->inversePowerC1 = inversePowerC1_i;
	pi_->inversePowerC2 = inversePowerC2_i;
}

vector<double> MCPacking::getTotalDisplacement(){
	return pi_->totalDisplacement;
}

void MCPacking::setTauAlpha(long tauAlpha_i){
	tauAlpha = tauAlpha_i;
}

double MCPacking::getPhi(){
	return pi_->phi;
}

double MCPacking::getAcceptanceRatio(){
	return acceptanceRatio;
}

double MCPacking::getLEshift(){
	return pi_->LEshift;
}

//set/get Constraint
void MCPacking::setConstraint(vector<double> conArray) {
	if(pi_->simControl.constraintType==simControlStruct::constraintEnum::strainControl){
		if (long(conArray.size()) == long(pi_->numParticles * pi_->nDim)) {
			for (long i = 0; i < pi_->nDim * pi_->numParticles; i++){
				pi_->constraint[i] = conArray[i];
			}
		} else {
			cout << "Constraint array is the wrong size" << endl;
		}
	} else {
		cout << "Can't set constraint outside of strainControl enum" << endl;
	}
}

void MCPacking::setLEshift(double LEData){
	if(pi_->simControl.geometryType==simControlStruct::geometryEnum::leesEdwards){
		pi_->LEshift = LEData;
	}else{
		cout << "Can't set LEshift outside of leesEdwards enum" << endl;
	}
}


vector<double> MCPacking::getConstraint() {
	return pi_->constraint;
}

double MCPacking::getConstraintMagnitude() {
	double constraintMag = 0.0;
	if(pi_->simControl.constraintType==simControlStruct::constraintEnum::strainControl){
		calcForceEnergy();
		constraintMag = pi_->forceDotConstraint;
	} else{
		cout << "Can't get constraint magnitude outside of strainControl enum" << endl;
	}
	return constraintMag;
}

void MCPacking::setExternalForce(vector<double> externalForceArray) {
	if(pi_->simControl.constraintType==simControlStruct::constraintEnum::stressControl){
		if (long(externalForceArray.size()) == long(pi_->numParticles * pi_->nDim)) {
			for (long i = 0; i < pi_->nDim * pi_->numParticles; i++){
				pi_->externalForce[i] = externalForceArray[i];
			}
			calcForceEnergy();
		} else {
			cout << "External force array is the wrong size" << endl;
		}
	} else {
		cout << "Can't set external force outside of stressControl enum" << endl;
	}
}

vector<double> MCPacking::getExternalForce() {
	return pi_->externalForce;
}

void MCPacking::calcCenterOfMass(){
	for (int j = 0; j < pi_->nDim; j++){
		pi_->centerOfMass[j] = 0;
		for (int i = 0; i < pi_->numParticles; i++){
			pi_->centerOfMass[j] += pi_->unwrappedPos[i*pi_->nDim + j];
		}
		pi_->centerOfMass[j] /= pi_->numParticles;
	}
}

void MCPacking::setPhi(double phi_i){
	pi_->calcCurrentPhi();
	double scalingFactor = pow((phi_i/pi_->phi),(1.0/pi_->nDim));
	for (long i = 0; i < pi_->numParticles; i++){
		pi_->radii[i] *= scalingFactor;
	}
	pi_->phi = phi_i;
}

void MCPacking::setNDim(int dim_i){
	pi_->nDim = dim_i;
}

vector<bool> MCPacking::uniqueParticles(double cutoff){
	vector<bool> toDelete(pi_->numParticles, false);
	for (long i = 0; i < pi_->numParticles - 1; i++){
		if(!toDelete[i]){
			for (long j = i+1; j < pi_->numParticles; j++){
				if (pbcDistance(pi_->pos.begin() + i*pi_->nDim,pi_->pos.begin() + j*pi_->nDim, i, j)<cutoff){
					toDelete[j] = true;
				}
			}
		}
	}
	return toDelete;
}

tuple<vector<long>, vector<long> > MCPacking::getNeighbors(){
	vector<long> neighborIndptr;
	vector<long> neighborIndices;
	pi_->doubleVectorToCSR(neighborList, neighborIndices, neighborIndptr);
	return make_tuple(neighborIndices, neighborIndptr);
}

//Set the list of nearby neighbors for each particle
// void MCPacking::setNeighbors(vector<long> indices, vector<long> indptr) {
// 	if((indptr.size == pi_->numParticles + 1) && (indices.size == indptr.data[pi_->numParticles])) {
// 		thrust::device_vector<long> d_indices;
// 		thrust::device_vector<long> d_indptr;
// 		pi_->intArrayInToDeviceVector(d_indices, indices);
// 		pi_->intArrayInToDeviceVector(d_indptr, indptr);
// 		neigh_->d_maxNeighborList.resize(pi_->numParticles);
// 		thrust::transform(d_indptr.begin() + 1, d_indptr.end(), d_indptr.begin(), neigh_->d_maxNeighborList.begin(), thrust::minus<long>());
// 		neigh_->maxNeighbors = thrust::reduce(thrust::cuda::par(pi_->alloc), neigh_->d_maxNeighborList.begin(), neigh_->d_maxNeighborList.end(), -1L, thrust::maximum<long>());
// 		neigh_->neighborListSize = pow(2, ceil(std::log2(neigh_->maxNeighbors)));
// 		neigh_->d_neighborList.resize(pi_->numParticles * neigh_->neighborListSize);
// 		thrust::fill(neigh_->d_neighborList.begin(), neigh_->d_neighborList.end(), -1);

// 		long* indicesPtr = thrust::raw_pointer_cast(&d_indices[0]);
// 		long* indptrPtr = thrust::raw_pointer_cast(&d_indptr[0]);
// 		long* maxNeighborListPtr = thrust::raw_pointer_cast(&neigh_->d_maxNeighborList[0]);
// 		long* neighborListPtr = thrust::raw_pointer_cast(&neigh_->d_neighborList[0]);
// 		long maxNeighbors = neigh_->maxNeighbors;
// 		long& neighborListSize = neigh_->neighborListSize;

// 		auto transfer = [=] __device__(long particleID) {
// 			long start = indptrPtr[particleID];
// 			for(long nID = 0; nID < maxNeighbors; nID++) {
// 				if(nID < maxNeighborListPtr[particleID]) {
// 					neighborListPtr[particleID * neighborListSize + nID] = indicesPtr[start + nID];
// 				}
// 			}
// 		};

// 		for (long pID = 0; pID < pi_->numParticles; pID++){

// 		}
// 		thrust::for_each(r, r + pi_->numParticles, transfer);
// 	} else {
// 		cout << "Neighbor list does not match numParticles" << endl;
// 	}
// }

//Returns the minimum gap among pairs of interacting particles
double MCPacking::getMinGap(double maxSearchGap) {
	vector<long> indices1;
	vector<long> indices2;
	vector<double> gapValues;
	tie(indices1, indices2, gapValues) = getSmallGaps(maxSearchGap);
	double minGap = *min_element(gapValues.begin(),gapValues.end());
	// for(size_t i = 0; i < gapValues.size(); i++){
	// 	if (gapValues[i] > 0.0 && gapValues[i] < minGap){
	// 		minGap = gapValues[i];
	// 	}
	// }
	return minGap;
	// This is the right way to do it
	// long numGaps;
	// switch(pi_->simControl.neighborType) {
	// 	case simControlStruct::neighborEnum::hex2D:
	// 		numGaps = 6;
	// 		break;
	// 	case simControlStruct::neighborEnum::allToAll:
	// 		cout << __PRETTY_FUNCTION__ << ": Warning: neighborType is allToAll, gapList is computed over all particles" << endl;
	// 		numGaps = pi_->numParticles;
	// 		break;
	// 	default:
	// 		numGaps = maxNeighbors;
	// 		break;
	// }
	// return pi_->hostnumToQuadArrayOut(hostnum(pi_->calcMinGap(numGaps)));
}

void MCPacking::resetCenterOfMass(){
	for (int i = 0; i < pi_->nDim; i++){
		pi_->centerOfMassShift[i] = 0.0;
	}
}

vector<double> MCPacking::createD9Plus(vector<double> crystalPos){
	long nP = crystalPos.size()/9;
	vector<bool> toDelete(nP,false);
	vector<double> firstPassPos(crystalPos.size(),0.0);
	long firstPassNum = 0;		

	cout << nP << endl;

	for(int32_t i = 0; i < nP; i++){
		cout << "1: " << i << endl;
		vector<int32_t> mirror(pi_->nDim,0);
		dnLatticePoint(crystalPos.begin() + i*9,mirror);
		for(int32_t j = 0; j < 9; j++){
			if (mirror[j]!=0){
				break;
			}
		}
		for(int32_t k = 0; k < 9; k++){
			firstPassPos[firstPassNum*9 + k] = crystalPos[i*9 + k];
		}
		firstPassNum++;
	}
	
	cout << firstPassNum << endl;
	
	firstPassPos.resize(firstPassNum*9);
	vector<bool> toDelete2(firstPassNum,false);
	double dist;
	double compare = 0.00001;
	for (int32_t i = 0; i < firstPassNum - 1; i++){
		if(!toDelete2[i]){
			cout << "2: " << i << endl;
			for (int32_t j = i+1; j < firstPassNum; j++){
				if(!toDelete2[j]){
					dist = pbcDistance(firstPassPos.begin() + i*9,firstPassPos.begin() + j*9, -1, -1);
					if(dist<compare){
						toDelete2[j] = false;
					}
				}
			}
		}
	}
	long secondPassNum = 0;
	vector<double> secondPassPos(firstPassNum*9,0.0);
	for (int32_t i = 0; i < firstPassNum; i++){
		if (!toDelete2[i]){
			for(int32_t j=0; j<9; j++){
				secondPassPos[secondPassNum*9 + j] = firstPassPos[i*9 + j];
			}
			secondPassNum++;
		}
	}
	secondPassPos.resize(firstPassNum*9);
	return secondPassPos;
}

void MCPacking::setNumParticles(long numParticles_i){
	pi_->numParticles = numParticles_i;
}

void MCPacking::setPotentialPower(double potentialPower_i){
	pi_->potentialPower = potentialPower_i;
}

double MCPacking::getPotentialPower(){
	return pi_->potentialPower;
}

double MCPacking::getVolume(){
	return pi_->volume;
}

//FIRE minimizer components
void MCPacking::initFIREMinimizer(vector<double> FIREParams_, double maxForceRescale_, long N_min_, long lastRestart_) {
	if (FIREParams_.size() == 7) {
    	double maxForceRescale = maxForceRescale_;
    	double a_start_ = FIREParams_[0];
		double f_dec_ = FIREParams_[1];
		double f_inc_ = FIREParams_[2];
		double f_a_ = FIREParams_[3];
		double dt_ = FIREParams_[4];
		double dt_max_ = FIREParams_[5];
		double a_ = FIREParams_[6];
		fire_->initMinimizer(a_start_, f_dec_, f_inc_, f_a_, dt_, dt_max_, maxForceRescale, N_min_, a_, lastRestart_);
	} else {
		cout << "Wrong number of parameters entered, must be 7" << endl;
	}
}

void MCPacking::minimizerLoopFIRE() {
	fire_->minimizerLoop();
}

void MCPacking::rescaleForces(double maxForceScale) {
	pi_->rescaleForces(maxForceScale);
}

void MCPacking::setMaxForceRescale(double maxForceScale) {
	pi_->maxForceRescale = maxForceScale;
}

//set/get number of iterations for FIRE
void MCPacking::setFIREIteration(long FIREIteration_) {
	fire_->lastRestart = FIREIteration_;
}

long MCPacking::getFIREIteration() {
	return fire_->lastRestart;
}

void MCPacking::setFIREDamping(double fire_alpha_){
	fire_->fire_alpha = fire_alpha_;
}

double MCPacking::getFIREDamping(){
	return fire_->fire_alpha;
}

double MCPacking::getFIRETimeStep(){
	return fire_->dt;
}

void MCPacking::setFIRETimeStep(double dt_){
	fire_->dt = dt_;
}

void MCPacking::resetFIREMaxTimeStep(double dt_max_){
	fire_->dt_max = dt_max_;
}

void MCPacking::zeroVelocities(){
	fill(fire_->velocity.begin(), fire_->velocity.end(), 0.0);
}

static bool abs_compare(double a, double b){
    return (std::abs(a) < std::abs(b));
}

double MCPacking::pbcDistance(vector<double>::iterator pos1, vector<double>::iterator pos2, long pID, long nID){
	return sqrt(pbcDistanceSq(pos1,pos2,pID,nID));
}

double MCPacking::pbcDistance(const double x, const double y) {
	double delta = x - y;
	return ( delta - rint(delta) );
}

//pbcDistance w/ non-square boundaries
double MCPacking::pbcDistance(const double x, const double y, const double size) {
	// devnum delta = (x - y)/size;
	// return ( delta - rint(delta) ) *size;
	//Implement pbcDistance without using rint or division
	//The cost is that this only works if we are within s of the boundary
	//Could be turned into a single function if we use default value of size = 1
	double delta = x-y;
	if (2*delta < -size) return delta + size;
	if (2*delta > size) return delta - size;
	return delta;
}

// Overloaded to take in vectors
double MCPacking::pbcDistanceOutside(vector<double> pos1, vector<double> pos2){
	return pbcDistance(pos1.begin(), pos2.begin(), -1, -1);
}

vector<double> MCPacking::pbcDisplacementOutside(vector<double> pos1, vector<double> pos2){
	return pbcDisplacement(pos1.begin(),pos2.begin(),-1,-1);
}

double MCPacking::pbcDistanceSq(vector<double>::iterator pos1, vector<double>::iterator pos2, long pID, long nID){
	double accum = 0.0;
	vector<double> dispVector = pbcDisplacement(pos1, pos2, pID, nID);
	for (int dimIndex = 0; dimIndex < pi_->nDim; dimIndex++){
		accum += (dispVector[dimIndex]*dispVector[dimIndex]);
	}
	return accum;
	// switch(pi_->simControl.particleType) {
	// 	case simControlStruct::particleEnum::sphere:{
	// 		for (int dimIndex = 0; dimIndex < pi_->nDim; dimIndex++){
	// 			accum += (dispVector[dimIndex]*dispVector[dimIndex]);
	// 		}
	// 		return accum;
	// 	}
	// 	case simControlStruct::particleEnum::cube:{
	// 		accum = *max_element(dispVector.begin(),dispVector.end(),abs_compare);
	// 		accum *= accum;
	// 	}
	// }
	// return accum;
}



vector<double> MCPacking::pbcDisplacement(vector<double>::iterator pos1, vector<double>::iterator pos2, long pID, long nID){
	vector<double> displacement(pi_->nDim);
	vector<double> pos2Copy(pi_->nDim);
	// pos2Copy = pos2;
	for(int dim=0; dim < pi_->nDim; dim++){
		pos2Copy[dim] = pos2[dim];
	}
	if ((pi_->simControl.distanceType==simControlStruct::distanceEnum::MK)&&(pID!=-1)&&(nID!=-1)){
		for(int dim=0; dim < pi_->nDim; dim++){
			pos2Copy[dim] += pi_->MKLambdas[nID*pi_->numParticles*pi_->nDim + pID*pi_->nDim + dim];
		}
	}
	switch(pi_->simControl.geometryType) {
		case simControlStruct::geometryEnum::nonSquare:
			for (int dimIndex = 0; dimIndex < pi_->nDim; dimIndex++){
				double delta = (pos1[dimIndex] - pos2Copy[dimIndex])/pi_->boxSize[dimIndex];
				displacement[dimIndex] = (delta - fastRound(delta))*pi_->boxSize[dimIndex];
			}
			break;
		case simControlStruct::geometryEnum::E8:{
				for (int dimIndex = 0; dimIndex < pi_->nDim; dimIndex++){
					displacement[dimIndex] = pos1[dimIndex] - pos2Copy[dimIndex];
				}
				vector<int32_t> mirror(pi_->nDim,0);
				e8LatticePoint(displacement.begin(),mirror);
			}
			break;
		case simControlStruct::geometryEnum::Dn:{
				for (int dimIndex = 0; dimIndex < pi_->nDim; dimIndex++){
					displacement[dimIndex] = pos1[dimIndex] - pos2Copy[dimIndex];
				}
				vector<int32_t> mirror(pi_->nDim,0);
				dnLatticePoint(displacement.begin(),mirror);
			}
			break;
		case simControlStruct::geometryEnum::leesEdwards:{
			for (int dimIndex = 0; dimIndex < pi_->nDim; dimIndex++){
				if (dimIndex == 1){
					double delta1 = (pos1[0] - pos2Copy[0]);
					// double rounded = rint(delta1); //need to store for lees-edwards BC
					double rounded = fastRound(delta1); //need to store for lees-edwards BC
					delta1 -= rounded;
					double delta2 = (pos1[1] - pos2Copy[1]);
					displacement[dimIndex] = delta2  - rounded*pi_->LEshift - rint(delta2 - rounded*pi_->LEshift);
				}else{
					double delta = pos1[dimIndex] - pos2Copy[dimIndex];
					displacement[dimIndex] = delta - fastRound(delta);
				}
			}
			break;
		}
		default:
			for (int dimIndex = 0; dimIndex < pi_->nDim; dimIndex++){
				double delta = pos1[dimIndex] - pos2Copy[dimIndex];
				displacement[dimIndex] = delta - fastRound(delta);
			}
			break;
	}
	return displacement;
}

void MCPacking::setMCMCBlockedDirection(vector<bool> mcmcBlockedDirection_i){
	for(int i=0; i<pi_->nDim; i++){
		mcmcBlockedDirection[i] = mcmcBlockedDirection_i[i];
	}
}

vector<bool> MCPacking::getMCMCBlockedDirection(){
	return mcmcBlockedDirection;
}

vector<double> MCPacking::getRandomVector(long numEntries, double maxElement){
	vector<double> randomVector(numEntries);
	for(long i = 0; i < numEntries; i++){
		randomVector[i] = uniformDist(mt)*maxElement;
	}
	return randomVector;
}

vector<double> MCPacking::getRandomGaussianVector(long numEntries, double mean, double std){
	vector<double> randomVector(numEntries);
	normal_distribution<double> distribution(mean,std);
	for(long i = 0; i < numEntries; i++){
		randomVector[i] = distribution(mt);
	}
	return randomVector;
}

double MCPacking::fastRound(double numToRound){
	double dtmp = numToRound + 6755399441055744.0;
	return reinterpret_cast<int&>(dtmp);
}

// float MCPacking::fastRound(float numToRound){
// 	float dtmp = numToRound + (float) 12582912;
// 	return reinterpret_cast<int16_t&>(dtmp);
// }

/* latticepoint calculates the Dn minimum image */
void MCPacking::dnLatticePoint(vector<double>::iterator p, vector<int32_t> &mirror){
	int rp, gind = 0, summr = 0;
	double maxdelta = 0.0, dtmp;
	for(rp=0; rp<pi_->nDim; ++rp){
	    // fast round eqv 
	    // mirror[rp] = round(p[rp]);
		// dtmp = p[rp] + 6755399441055744.0;
		// mirror[rp] = reinterpret_cast<int32_t&>(dtmp);
		mirror[rp] = fastRound(p[rp]);
		summr += mirror[rp];
		p[rp] -= mirror[rp];
	}
	if(summr%2 != 0){
		for(rp=0; rp<pi_->nDim; ++rp){
			dtmp = fabs(p[rp]);
			if(maxdelta < dtmp){
				maxdelta = dtmp;
				gind = rp;
			}
		}
		// use next nearest neighbor image
		if(p[gind] < 0){
			--mirror[gind];
			++p[gind];
		} else {
			++mirror[gind];
			--p[gind];
		}
	}
}

/**
 * Minimum image decomposition of the vector
 * @f[ p_mathrm{in} = p_\mathrm{out} + \mathrm{mirror} @f]
 * @fD+n@f lattice is a coset of two @fDn@f lattices 
 * with an offset=(1/2) in all coordinates; valid in even dimension
* @fD0+n@f lattice is a coset of two @fDn@f lattices 
 * with an offset=(1/2, ..., 1/2, 0); valid in odd dimension
 *
 * @param[in, out] p DIM-dimensional vector of doubles. Get overwrote 
 * @param[out] mirror store the lattice vector. Initialization not necessary
 * 
 * @return int. 0 - mirror is as it is; 2 - mirror should add the offset
 */ 
int MCPacking::e8LatticePoint(vector<double>::iterator p, vector<int32_t> &mirror){
	int rp, rst = 0;
	double norm1=0.0, norm2 = 0.0;
	vector<double> pb(pi_->nDim);
	vector<int32_t> mirrorb(pi_->nDim);
	if (pi_->nDim%2 == 0){
		// even dimension
		for(rp=0; rp<pi_->nDim; ++rp){
			pb[rp] = p[rp] - 0.5;
		}
	} else {
		// odd dimension
		// pb[pi_->nDim-1] = p[pi_->nDim-1];
		for(rp=0; rp<pi_->nDim-1; ++rp){
		// for(rp=0; rp<pi_->nDim; ++rp){
			pb[rp] = p[rp] - 0.5;
		}
	}
	dnLatticePoint(p, mirror);
	dnLatticePoint(pb.begin(), mirrorb);
	for(rp=0; rp<pi_->nDim; ++rp){
		norm1 += p[rp]*p[rp];
		norm2 += pb[rp]*pb[rp];
	}
	if(norm2<norm1){
		// copy pb to p
		for(rp=0; rp<pi_->nDim; ++rp){
			p[rp] = pb[rp];
			mirror[rp] = mirrorb[rp];
		}
		rst = 2; // shift (1/2)
	}
	return rst;
}

// Should refactor with pbcDistance
vector<double> MCPacking::calcNVector(const long pID, const long nID){	
	double accum = 0.0;
	// vector<double> nVector(pi_->nDim);
	// vector<double> pos1(pi_->pos.begin() + pID * pi_->nDim, pi_->pos.begin() + (pID + 1) * pi_->nDim);
	// vector<double> pos2(pi_->pos.begin() + nID * pi_->nDim, pi_->pos.begin() + (nID + 1) * pi_->nDim);
	vector<double> nVector = pbcDisplacement(pi_->pos.begin() + nID * pi_->nDim, pi_->pos.begin() + pID * pi_->nDim, nID, pID);
	for (int dimIndex = 0; dimIndex < pi_->nDim; dimIndex++){
		// nVector[dimIndex] = pi_->pos[nID * pi_->nDim + dimIndex] - pi_->pos[pID * pi_->nDim + dimIndex];
		// nVector[dimIndex] -= round(nVector[dimIndex])*pi_->boxSize[dimIndex];
		accum += (nVector[dimIndex]*nVector[dimIndex]);
	}
	accum = sqrt(accum);
	for (int dimIndex = 0; dimIndex < pi_->nDim; dimIndex++){
		nVector[dimIndex] = nVector[dimIndex]/accum;
	}
	// transform(nVector.begin(),nVector.end(),nVector.begin(),bind(divides<double>(),placeholders::_1,sqrt(accum)));
	return nVector;
}

void MCPacking::setNeighborType(simControlStruct::neighborEnum neighborType_ ) {
	pi_->simControl.neighborType = neighborType_;
	// neigh_->initNeighbors();
}

simControlStruct::neighborEnum MCPacking::getNeighborType() {
	return pi_->simControl.neighborType;
}

void MCPacking::setDistanceType(simControlStruct::distanceEnum distanceType_ ) {
	pi_->simControl.distanceType = distanceType_;
}

simControlStruct::distanceEnum MCPacking::getDistanceType() {
	return pi_->simControl.distanceType;
}

void MCPacking::setGeometryType( simControlStruct::geometryEnum geometryType_ ) {
	pi_->simControl.geometryType = geometryType_;
	switch(pi_->simControl.geometryType) {
		case simControlStruct::geometryEnum::E8:{
				vector<double> newBoxSize(pi_->nDim,2.0);
				setBoxSize(newBoxSize);
				pi_->volume = 1.0;
			}
			break;
		case simControlStruct::geometryEnum::Dn:{
				double oldPhi = pi_->phi;
				vector<double> newBoxSize(pi_->nDim,2.0);
				setBoxSize(newBoxSize);
				pi_->volume = 2.0;
				setPhi(oldPhi);
			}
			break;
		default:
			pi_->volume = 1.0;
			break;
	}
}

simControlStruct::geometryEnum MCPacking::getGeometryType() {
	return pi_->simControl.geometryType;
}

void MCPacking::setPotentialType( simControlStruct::potentialEnum potentialType_ ) {
	pi_->simControl.potentialType = potentialType_;
}

simControlStruct::potentialEnum MCPacking::getPotentialType() {
	return pi_->simControl.potentialType;
}

void MCPacking::setParticleType(simControlStruct::particleEnum particleType_){
	pi_->simControl.particleType = particleType_;
}

simControlStruct::particleEnum MCPacking::getParticleType(){
	return pi_->simControl.particleType;
}

void MCPacking::setMonteCarloType(simControlStruct::monteCarloEnum monteCarloType_){
	pi_->simControl.monteCarloType = monteCarloType_;
}

simControlStruct::monteCarloEnum MCPacking::getMonteCarloType(){
	return pi_->simControl.monteCarloType;
}

void MCPacking::setEinsteinType( simControlStruct::einsteinEnum einsteinType_ ) {
	pi_->simControl.einsteinType = einsteinType_;
}

simControlStruct::einsteinEnum MCPacking::getEinsteinType() {
	return pi_->simControl.einsteinType;
}

void MCPacking::setEnsembleType(simControlStruct::ensembleEnum ensembleType_){
	pi_->simControl.ensembleType = ensembleType_;
}

simControlStruct::ensembleEnum MCPacking::getEnsembleType(){
	return pi_->simControl.ensembleType;
}

void MCPacking::setConstraintType(simControlStruct::constraintEnum constraintType_ ) {
	pi_->simControl.constraintType = constraintType_;
	// neigh_->initNeighbors();
}

simControlStruct::constraintEnum MCPacking::getConstraintType() {
	return pi_->simControl.constraintType;
}

void MCPacking::setLatticeSpacing(double latticeSpacing_) {
	pi_->latticeSpacing = latticeSpacing_;
	// neigh_->initNeighbors();
}

double MCPacking::getLatticeSpacing() {
	return pi_->latticeSpacing;
}

double MCPacking::getMeanRadius(){
	double accum = 0.0;
	for (int i = 0; i < pi_->numParticles; i++){
		accum += pi_->radii[i];
	}
	return accum/pi_->numParticles;
}

void MCPacking::removeParticles(vector<long> pIDList){
	for (size_t i = 0; i < pIDList.size(); i++){
		long pID = pIDList[i];
		cout << "Removing " << pID << endl;

		// TODO URGENT: Deal with MK lambdas and initialPositions 
		pi_->radii.erase(pi_->radii.begin() + pID);
		// cout << "here" << endl;
		pi_->distTravelledSq.erase(pi_->distTravelledSq.begin() + pID);
		// cout << "here1" << endl;
		energyVector.erase(energyVector.begin() + pID);
		// cout << "here2" << endl;
		pi_->pos.erase(pi_->pos.begin() + pID * pi_->nDim, pi_->pos.begin() + (pID + 1) * pi_->nDim);
		// cout << "here3" << endl;
		pi_->unwrappedPos.erase(pi_->unwrappedPos.begin() + pID * pi_->nDim, pi_->unwrappedPos.begin() + (pID + 1) * pi_->nDim);
		// cout << "here4" << endl;
		pi_->crystalPos.erase(pi_->crystalPos.begin() + pID * pi_->nDim, pi_->crystalPos.begin() + (pID + 1) * pi_->nDim);
		// cout << "here5" << endl;
		pi_->externalForce.erase(pi_->externalForce.begin() + pID * pi_->nDim, pi_->externalForce.begin() + (pID + 1) * pi_->nDim);
		// cout << "here6" << endl;
		pi_->constraint.erase(pi_->constraint.begin() + pID * pi_->nDim, pi_->constraint.begin() + (pID + 1) * pi_->nDim);
		// cout << "here7" << endl;
		pi_->lastPos.erase(pi_->lastPos.begin() + pID * pi_->nDim, pi_->lastPos.begin() + (pID + 1) * pi_->nDim);
		// cout << "here8" << endl;
		pi_->totalDisplacement.erase(pi_->totalDisplacement.begin() + pID * pi_->nDim, pi_->totalDisplacement.begin() + (pID + 1) * pi_->nDim);
		// cout << "here9" << endl;
		pi_->netForce.erase(pi_->netForce.begin() + pID * pi_->nDim, pi_->netForce.begin() + (pID + 1) * pi_->nDim);
		// initialPositions.erase(initialPositions.begin() + pID * pi_->nDim, initialPositions.begin() + (pID + 1) * pi_->nDim);
		// cout << "here" << endl;
		for (long j = 0; j < pi_->numParticles; j++){
			for(size_t k = neighborList[j].size(); k > 0 ; k--){
				if(neighborList[j][k-1] > pID){
					neighborList[j][k-1]--;
				}else if(neighborList[j][k-1] == pID){
					neighborList[j].erase(neighborList[j].begin() + k - 1);
				}
			}
		}
		neighborList.erase(neighborList.begin() + pID);
		pi_->numParticles--;
		pi_->calcCurrentPhi();
		// cout << "Removed!" << endl;
	}


	// Reset neighborList

	// Recalculate phi
	// contactList.resize(pi_->numParticles);
	// neighborList.resize(pi_->numParticles);

	// setPhi(phi_i);
	// initForceEnergy();

	// calcForceEnergy();
	resetCenterOfMass();
}

void MCPacking::setPositions(vector<double> newPos){
	if(int(newPos.size()) == pi_->numParticles * pi_->nDim){
		// Make sure that these fit into the periodic box
		switch(pi_->simControl.geometryType) {
			case simControlStruct::geometryEnum::nonSquare:
				for (int i = 0; i < pi_->nDim * pi_->numParticles; i++){
					pi_->pos[i] = newPos[i] - pi_->boxSize[i%pi_->nDim]*floor(newPos[i]/pi_->boxSize[i%pi_->nDim]);
					pi_->unwrappedPos[i] = newPos[i];
				}
				break;
			case simControlStruct::geometryEnum::E8:{
					vector<double> thisPos(pi_->nDim);
					vector<int32_t> mirror(pi_->nDim,0);
					for (long pID = 0; pID < pi_->numParticles; pID++){
						for (int dimIndex = 0; dimIndex < pi_->nDim; dimIndex++){
							thisPos[dimIndex] = newPos[pID * pi_->nDim + dimIndex];
						}
						e8LatticePoint(thisPos.begin(),mirror);
						for (int dimIndex = 0; dimIndex < pi_->nDim; dimIndex++){
							pi_->pos[pID * pi_->nDim + dimIndex] = thisPos[dimIndex];
						}
					}
					for (int i = 0; i < pi_->nDim * pi_->numParticles; i++){
						pi_->unwrappedPos[i] = newPos[i];
					}
				}
				break;
			case simControlStruct::geometryEnum::Dn:{
					vector<double> thisPos(pi_->nDim);
					vector<int32_t> mirror(pi_->nDim,0);
					for (long pID = 0; pID < pi_->numParticles; pID++){
						for (int dimIndex = 0; dimIndex < pi_->nDim; dimIndex++){
							thisPos[dimIndex] = newPos[pID * pi_->nDim + dimIndex];
						}
						dnLatticePoint(thisPos.begin(),mirror);
						for (int dimIndex = 0; dimIndex < pi_->nDim; dimIndex++){
							pi_->pos[pID * pi_->nDim + dimIndex] = thisPos[dimIndex];
						}
					}
					for (int i = 0; i < pi_->nDim * pi_->numParticles; i++){
						pi_->unwrappedPos[i] = newPos[i];
					}
				}
				break;
			default:
				for (int i = 0; i < pi_->nDim * pi_->numParticles; i++){
					pi_->pos[i] = newPos[i] - floor(newPos[i]);
					pi_->unwrappedPos[i] = newPos[i];
				}
				break;
		}
	}else{
		cout << "pos needs to have [numParticles * dim] entries" << endl;
	}
}

void MCPacking::setBoxSize(vector<double> newBoxSize){
	//TODO: Make sure that the user doesn't specify inf,nan,negative, or zero side lengths
	if(int(newBoxSize.size()) == pi_->nDim){
		if(pi_->simControl.geometryType == simControlStruct::geometryEnum::E8 || pi_->simControl.geometryType == simControlStruct::geometryEnum::Dn){
			// TODO: Scale E8 properly. For now, it must be in a 2x2x...x2 box
			for (int i = 0; i < pi_->nDim; i++){
				pi_->boxSize[i] = 2.0;
			}
		} else {
			// First we save the positions. We will simply set them at the end
			// and let setPositions take care of the boundary conditions
			vector<double> oldPos = getPositions();
			// We want to keep the volume the same when setting side lengths
			double volumeRoot = 1.0;
			for (int i = 0; i < pi_->nDim; i++){
				volumeRoot *= newBoxSize[i];
			}
			volumeRoot = pow(volumeRoot/pi_->volume,1.0/pi_->nDim);
			for (int i = 0; i < pi_->nDim; i++){
				pi_->boxSize[i] = newBoxSize[i]/volumeRoot;
			}
			// Now check to see if the box is square or not
			bool isSquare = true;
			for (int i = 1; i < pi_->nDim; i++){
				if (newBoxSize[i] != newBoxSize[0]){
					isSquare = false;
				}
			}
			// Set the geometry type automatically
			if (isSquare){
				setGeometryType(simControlStruct::geometryEnum::normal);
			} else {
				setGeometryType(simControlStruct::geometryEnum::nonSquare);
			}
			setPositions(oldPos);
		}
	}else{
		cout << "boxSize needs to have [dim] entries" << endl;
	}
}

void MCPacking::setCrystalPositions(vector<double> crystalPos_i){
	// These should be given as the unwrapped positions, just to keep
	// things organized. They will thus not be wrapped.
	if(int(crystalPos_i.size()) == pi_->numParticles * pi_->nDim){
		// Make sure that these fit into the periodic box
		for (int i = 0; i < pi_->nDim * pi_->numParticles; i++){
			pi_->crystalPos[i] = crystalPos_i[i];
		}
	}else{
		cout << "crystalPos needs to have [numParticles * dim] entries" << endl;
	}
}

void MCPacking::setRadii(vector<double> newRadii){
	if(int(newRadii.size()) == pi_->numParticles){
		for (int i = 0; i < pi_->numParticles; i++){
			pi_->radii[i] = newRadii[i];
		}
		pi_->calcCurrentPhi();
	}else{
		cout << "radii needs to have [numParticles] entries" << endl;
	}
}

void MCPacking::setD9Shift(vector<double> d9Shift_i){
	for(int i = 0; i < 9; i++){
		d9Shift[i] = d9Shift_i[i];
	}
}

void MCPacking::setMKLambdas(vector<double> MKLambdas_i){
	if(pi_->simControl.distanceType == simControlStruct::distanceEnum::MK){
		if(int(MKLambdas_i.size()) == pi_->numParticles * pi_->numParticles * pi_->nDim){
			pi_->MKLambdas.resize(pi_->numParticles*pi_->numParticles*pi_->nDim);
			for (int i = 0; i < pi_->numParticles*pi_->numParticles*pi_->nDim; i++){
				pi_->MKLambdas[i] = MKLambdas_i[i];
			}
		}else{
			cout << "MKLambdas needs to have [numParticles*numParticles*nDim] entries" << endl;
		}
	}else{
		cout << "Must use MK distanceType to set lambdas" << endl;
	}
}

void MCPacking::setMCMaxDisplacement(double maxDisplacement_i){
	maxDisplacement = maxDisplacement_i;
}

double MCPacking::getMCMaxDisplacement(){
	return maxDisplacement;
}

void MCPacking::setMCMaxDisplacementD9Relative(double maxDisplacementD9Relative_i){
	maxDisplacementD9Relative = maxDisplacementD9Relative_i;
}

double MCPacking::getMCMaxDisplacementD9Relative(){
	return maxDisplacementD9Relative;
}

void MCPacking::setMCMaxDisplacementCM(double maxDisplacementCM_i){
	maxDisplacementCM = maxDisplacementCM_i;
}

double MCPacking::getMCMaxDisplacementCM(){
	return maxDisplacementCM;
}

void MCPacking::setMaxVolumeShift(double maxVolumeShift_i){
	maxVolumeShift = maxVolumeShift_i;
}

double MCPacking::getMaxVolumeShift(){
	return maxVolumeShift;
}

double MCPacking::getTemperature(){
	return temperature;
}

void MCPacking::setTemperature(double temperature_i){
	temperature = temperature_i;
}

double MCPacking::getEinsteinLambda(){
	return einsteinLambda;
}

void MCPacking::setEinsteinLambda(double einsteinLambda_i){
	einsteinLambda = einsteinLambda_i;
}


vector<long> MCPacking::getRandomSequence(){
	vector<long> sequence(pi_->numParticles);
	iota(sequence.begin(),sequence.end(),0);
	std::shuffle(sequence.begin(),sequence.end(),mt);
	return sequence;
}

vector<double> MCPacking::getPositions(){
	return pi_->pos;
}

vector<double> MCPacking::getUnwrappedPositions(){
	return pi_->unwrappedPos;
}

vector<double> MCPacking::getCrystalPositions(){
	return pi_->crystalPos;
}

vector<double> MCPacking::getBoxSize(){
	return pi_->boxSize;
}

vector<double> MCPacking::getRadii(){
	return pi_->radii;
}

vector<double> MCPacking::getD9Shift(){
	return d9Shift;
}

vector<double> MCPacking::getMKLambdas(){
	return pi_->MKLambdas;
}

// tuple<vector<long>, vector<long>, vector<double> > MCPacking::getHessian(){
// 	long maxContacts = 0;
// 	vector<double> hessian;
// 	vector<long> contactList;
// 	vector<long> numContacts;
// 	switch (pi_->simControl.potentialType) {
// 		case simControlStruct::potentialEnum::contactPower:
// 			//Fall through to the same calculation as wca
// 		case simControlStruct::potentialEnum::inversePower:
// 			//Fall through to the same calculation as wca
// 		case simControlStruct::potentialEnum::wca:
// 			//Calculate the set of contacts
// 			// calcContacts(contactList, numContacts, 0);
// 			calcContacts();
// 			//Calculate the maximum number of contacts
// 			long maxContacts = *max_element(numContacts.begin(),numContacts.end());
// 			break;
// 		case simControlStruct::potentialEnum::log:
// 			//neigh_->updateNeighbors();
// 			//Set the list of number of contacts to the list of number of neighbors
// 			contactList = neigh_->neighborList;
// 			numContacts = neigh_->maxNeighborList;
// 			maxContacts = neigh_->maxNeighbors;
// 			break;
// 	}

// //	pi_->calcHessian(hessian, contactList, numContacts);
// 	vector<long> indices1;
// 	vector<long> indices2;
// 	vector<double> data;
// //	pi_->diagonalPaddedToCOO(contactList, numContacts, hessian, indices1, indices2, data, pi_->nDim, pi_->nDim);
// 	return make_tuple(indices1, indices2, data);
// }


bool MCPacking::singleMCHardSphere(const long pID, const vector<double>& translationVector){
	vector<double> translatedPos(pi_->nDim);
	for (int i = 0; i < pi_->nDim; i++){
		translatedPos[i] = pi_->unwrappedPos[pID*pi_->nDim + i] + translationVector[i];
	}
	// if(pi_->simControl.geometryType == simControlStruct::geometryEnum::Dn){
	// 	vector<int32_t> mirror(pi_->nDim,0);
	// 	dnLatticePoint(translatedPos.begin(),mirror);
	// }
	for(size_t i = 0; i < neighborList[pID].size(); i++){
		long nID = neighborList[pID][i];
		if (calcOverlap(translatedPos.begin(),pi_->unwrappedPos.begin() + nID * pi_->nDim, pID, nID) > 0.0) {
			return false;
		}
	}
	pi_->translateParticle(pID,translationVector);
	shiftCenterOfMass(translationVector,pi_->numParticles);
	return true;
}

void MCPacking::translate(const double stepSize, vector<double> translationVector){
	if(int(translationVector.size()) == pi_->numParticles * pi_->nDim){
		for(long i = 0; i < pi_->numParticles * pi_->nDim; i++){
			translationVector[i] *= stepSize;
		}
		for(long i = 0; i < pi_->numParticles; i++){
			vector<double> thisVector(translationVector.begin() + i * pi_->nDim,translationVector.begin() + (i+1) * pi_->nDim);
			pi_->translateParticle(i,thisVector);
		}
	}else{
		cout << "translation vector needs to have [numParticles * dim] entries" << endl;
	}
}


bool MCPacking::singleMCThermal(const long pID, const vector<double>& translationVector){
	bool makeMove = true;
	vector<double> translatedPos(pi_->nDim);
	double movedEnergy = 0.0;
	for (int i = 0; i < pi_->nDim; i++){
		translatedPos[i] = pi_->unwrappedPos[pID*pi_->nDim + i] + translationVector[i];
	}
	for(size_t i = 0; i < neighborList[pID].size(); i++){
		long nID = neighborList[pID][i];
		double thisDist = pbcDistanceSq(translatedPos.begin(),pi_->unwrappedPos.begin() + nID * pi_->nDim, pID, nID);
		// TODO: Test if the if statement is necessary
		if (thisDist < (getRadSum(pi_->radii[pID], pi_->radii[nID]))) {
			movedEnergy += calcEnergyPerInteraction(translatedPos.begin(),pi_->unwrappedPos.begin() + nID * pi_->nDim, pID, nID);
		}
	}
	makeMove = min(exp((energyVector[pID] - movedEnergy)/double(temperature)),1.0) > uniformDist(mt);
	if (makeMove) {
		pi_->translateParticle(pID,translationVector);
		shiftCenterOfMass(translationVector,pi_->numParticles);
	}
	return makeMove;
}

bool MCPacking::MCStepVolume(){
	// TODO: Add non-HS cases
	double initialPhi = pi_->phi;
	double minRadius = *min_element(pi_->radii.begin(),pi_->radii.end());
	double currentVolume = pi_->volume / halfPowerPow(2.0 * minRadius,pi_->nDim);
	double trialVolume = exp(log(currentVolume) + maxVolumeShift * (uniformDist(mt) - 0.5));
	// double trialVolume = currentVolume + maxVolumeShift * (uniformDist(mt) - 0.5) / halfPowerPow(2.0 * minRadius,pi_->nDim);
	double trialPhi = pi_->phi * currentVolume / trialVolume;
	// cout << "currentVolume is " << currentVolume << endl;
	// cout << "trialVolume is " << trialVolume << endl;
	// cout << "trialPhi is " << trialPhi << endl;
	// double boltzmannFactor = exp(-(pi_->targetPressure * (trialVolume - currentVolume) - (pi_->numParticles) * log(trialVolume/currentVolume)));
	double boltzmannFactor = exp(-(pi_->targetPressure * (trialVolume - currentVolume) - (pi_->numParticles+1) * log(trialVolume/currentVolume)));
	double randomNumber = uniformDist(mt);
	bool makeMove = boltzmannFactor > randomNumber;
	if (makeMove) {
		setPhi(trialPhi);
		if (anyOverlap()){
			setPhi(initialPhi);
			makeMove = false;
		}
	}
	return makeMove;
}

bool MCPacking::MCStepVolumeStretch(int direction){
	// TODO: Add non-HS cases
	double initialPhi = pi_->phi;
	double minRadius = *min_element(pi_->radii.begin(),pi_->radii.end());
	double currentVolume = pi_->volume / halfPowerPow(2.0 * minRadius,pi_->nDim);
	double trialVolume = exp(log(currentVolume) + maxVolumeShift * (uniformDist(mt) - 0.5));
	// double trialVolume = currentVolume + maxVolumeShift * (uniformDist(mt) - 0.5) / halfPowerPow(2.0 * minRadius,pi_->nDim);
	double trialPhi = pi_->phi * currentVolume / trialVolume;
	// cout << "currentVolume is " << currentVolume << endl;
	// cout << "trialVolume is " << trialVolume << endl;
	// cout << "trialPhi is " << trialPhi << endl;
	// double boltzmannFactor = exp(-(pi_->targetPressure * (trialVolume - currentVolume) - (pi_->numParticles) * log(trialVolume/currentVolume)));
	double boltzmannFactor = exp(-(pi_->targetPressure * (trialVolume - currentVolume) - (pi_->numParticles+1) * log(trialVolume/currentVolume)));
	double randomNumber = uniformDist(mt);
	bool makeMove = boltzmannFactor > randomNumber;
	vector<double> initialBox = getBoxSize();
	vector<double> trialBox = initialBox;
	trialBox[direction] *= pow(trialVolume / currentVolume, 1.0 / double(pi_->nDim));
	if (makeMove) {
		setBoxSize(trialBox);
		setPhi(trialPhi);
		if (anyOverlap()){
			setBoxSize(initialBox);
			setPhi(initialPhi);
			makeMove = false;
		}
	}
	return makeMove;
}

bool MCPacking::MCStepAspectRatio(int direction){
	// TODO: Add non-HS cases
	double minRadius = *min_element(pi_->radii.begin(),pi_->radii.end());
	double currentVolume = pi_->volume / halfPowerPow(2.0 * minRadius,pi_->nDim);
	double trialVolume = exp(log(currentVolume) + maxVolumeShift * (uniformDist(mt) - 0.5));
	vector<double> initialPos = pi_->unwrappedPos;
	vector<double> initialBox = getBoxSize();
	vector<double> trialBox = initialBox;
	trialBox[direction] *= pow(double(trialVolume) / currentVolume, 1.0 / double(pi_->nDim));
	setBoxSize(trialBox);
	calcNeighborsCut(pi_->lastNeighborCut);
	if (anyOverlap()){
		setBoxSize(initialBox);
		setPositions(initialPos);
		return false;
	} else {
		return true;
	}
}

bool MCPacking::MCStepD9Relative(int dim, double translation){
	// Always move one set relative to the other. Split difference to keep com same 
	translation = translation/2.0;
	vector<double> translationVector(pi_->nDim, 0.0);
	translationVector[dim] = translation;
	vector<double> translatedPos(pi_->nDim * pi_->numParticles);
	bool makeMove = true;
	for (long pID = 0; pID < pi_->numParticles; pID++){
		for (int dimIndex = 0; dimIndex < pi_->nDim; dimIndex++){
			if (pID < 19683){
				translatedPos[pID*pi_->nDim + dimIndex] = pi_->unwrappedPos[pID*pi_->nDim + dimIndex] + translationVector[dimIndex];
			} else{
				translatedPos[pID*pi_->nDim + dimIndex] = pi_->unwrappedPos[pID*pi_->nDim + dimIndex] - translationVector[dimIndex];
			}
		}
	}

	// Check if there's a collision
	for (long pID = 0; pID < pi_->numParticles; pID++){
		for(size_t i = 0; i < neighborList[pID].size(); i++){
			long nID = neighborList[pID][i];
			if(nID > pID){
				double radSum = getRadSum(pi_->radii[pID],pi_->radii[nID]);
				if (pbcDistanceSq(translatedPos.begin() + pID * pi_->nDim,translatedPos.begin() + nID * pi_->nDim, pID, nID) < radSum*radSum) {
					return false;
				}
			}
		}
	}

	// Check if allowed by einstein energetics
	if (einsteinLambda > 0.0){
		double boltzmannFactor;
		double factor = 0.0;
		switch(pi_->simControl.einsteinType) {
			case simControlStruct::einsteinEnum::periodicUmbrellaExp:
				{
					boltzmannFactor = 0.5;
				}	
				break;
			case simControlStruct::einsteinEnum::periodic:
				{
					double ksp = 2.0 * PI / pi_->latticeSpacing;
					double normalization = getPeriodicRefNormalization();
					double lamNorm = getLamNorm();
					for (long pID = 0; pID < pi_->numParticles; pID++){
						double thisSign;
						vector<double> thisTranslation(pi_->nDim, 0.0);
						if (pID < 19683){
							thisTranslation[dim] = translation;
						}else{
							thisTranslation[dim] = -translation;
						}
						double cosListCurrent = calcPeriodicCosList(pID, ksp, vector<double>(pi_->nDim,0.0));
						double cosListNew = calcPeriodicCosList(pID, ksp, thisTranslation);
						factor += (cosListNew - cosListCurrent)/normalization;
					}
					if (pi_->nDim==9){
						// Factor of 2 from occupancy
						// factor = factor / 2.0;
					}
					boltzmannFactor = exp(-einsteinLambda*factor/lamNorm);
				}
				break;
			default:
				// TODO: Figure out what the generic case will look like
				break;
			}
		double randomNumber = uniformDist(mt);
		makeMove = boltzmannFactor > randomNumber;
	}

	if (makeMove){
		// Actually translate
		for (long pID = 0; pID < 19683; pID++){
			pi_->translateParticle(pID,translationVector);
		}
		translationVector[dim] = -translation;
		for (long pID = 19683; pID < pi_->numParticles; pID++){
			pi_->translateParticle(pID,translationVector);
		}
		d9Shift[dim] += translation;
	}
	return makeMove;
}

bool MCPacking::MCStepCM(){
	// Should only be used for outside access
	vector<double> translationVector(pi_->nDim);
	double minRadius = *min_element(pi_->radii.begin(),pi_->radii.end());
	for (int j = 0; j < pi_->nDim; j++){
		translationVector[j] = 2.0 * maxDisplacementCM * pi_->volume * minRadius * (uniformDist(mt) - 0.5);
	}
	return centerOfMassMoveMC(translationVector);
}

bool MCPacking::centerOfMassMoveMC(const vector<double>& translationVector){
	// TODO: Refactor with singleMoveEinstein
	bool makeMove = false;
	double factor = 0.0;
	double boltzmannFactor = 0.0;
	switch(pi_->simControl.einsteinType) {
		case simControlStruct::einsteinEnum::spring:
			return false;
		case simControlStruct::einsteinEnum::springUntethered:
			{
				double lamNorm = getLamNorm();
				for (int pID = 0; pID < pi_->numParticles; pID++){
					for (int j = 0; j < pi_->nDim; j++){
						factor += (2.0*(pi_->unwrappedPos[pID*pi_->nDim + j] - pi_->crystalPos[pID*pi_->nDim + j]) + translationVector[j]) * translationVector[j];
					}
				}
				boltzmannFactor = exp(-einsteinLambda*factor/lamNorm);
			}
			break;
		case simControlStruct::einsteinEnum::periodicUmbrella:
			{
				double ksp = 2.0 * PI / pi_->latticeSpacing;
				double normalization = getPeriodicRefNormalization();
				double lamNorm = getLamNorm();
				double wfactor = 0;
				for (long pID = 0; pID < pi_->numParticles; pID++){
					double cosListCurrent = calcPeriodicCosList(pID, ksp, vector<double>(pi_->nDim,0.0));
					double cosListNew = calcPeriodicCosList(pID, ksp, translationVector);
					factor += (cosListCurrent - cosListNew)/normalization;
					wfactor += (1.0 + cosListNew)/(1.0 + cosListCurrent);
					// wfactor += (1.0 + cosListNew)/(1.0 + cosListCurrent);
					// factor += (0.5*(1.0 - cosListNew) - 0.5*(1.0 - cosListCurrent));
				}
				// wfactor = 1.0;
				boltzmannFactor = wfactor*exp(-(einsteinLambda/lamNorm)*factor);
				// boltzmannFactor = 0.5;
			}	
			break;
		case simControlStruct::einsteinEnum::periodicUmbrellaExp:
			{
				boltzmannFactor = 0.5;
			}	
			break;
		case simControlStruct::einsteinEnum::periodic:
			{
				double ksp = 2.0 * PI / pi_->latticeSpacing;
				double normalization = getPeriodicRefNormalization();
				double lamNorm = getLamNorm();
				for (long pID = 0; pID < pi_->numParticles; pID++){
					double cosListCurrent = calcPeriodicCosList(pID, ksp, vector<double>(pi_->nDim,0.0));
					double cosListNew = calcPeriodicCosList(pID, ksp, translationVector);
					factor += (cosListNew - cosListCurrent)/normalization;
					// factor += ((1.0 - cosListNew) - (1.0 - cosListCurrent))/normalization;
					// factor += (0.5*(1.0 - cosListNew) - 0.5*(1.0 - cosListCurrent));
				}
				if (pi_->nDim==9){
					// Factor of 2 from occupancy
					// factor = factor / 2.0;
				}
				boltzmannFactor = exp(-einsteinLambda*factor/lamNorm);
			}
			break;
		default:
			// TODO: Figure out what the generic case will look like
			break;
		}

	// double boltzmannFactor = exp(-einsteinLambda*factor);
	double randomNumber = uniformDist(mt);
	makeMove = boltzmannFactor > randomNumber;
	if (makeMove) {
		for (long pID = 0; pID < pi_->numParticles; pID++){
			pi_->translateParticle(pID,translationVector);
		}
		shiftCenterOfMass(translationVector,1);
	}
	return makeMove;
}

double MCPacking::getEinsteinEnergy(){
	double energy = 0;
	double meanRadius = getMeanRadius();
	switch(pi_->simControl.einsteinType) {
		case simControlStruct::einsteinEnum::spring:
			{				
				for (long pID = 0; pID < pi_->numParticles; pID++){
					for (int j = 0; j < pi_->nDim; j++){
						double delta = (pi_->unwrappedPos[pID*pi_->nDim + j] - pi_->crystalPos[pID*pi_->nDim + j] - pi_->centerOfMassShift[j]);
						energy += delta*delta/(4.0 * meanRadius * meanRadius);
					}
				}
				energy *= einsteinLambda;
			}
			break;
		case simControlStruct::einsteinEnum::springUntethered:
			{	
				double lamNorm = (4.0 * meanRadius * meanRadius);
				for (long pID = 0; pID < pi_->numParticles; pID++){
					for (int j = 0; j < pi_->nDim; j++){
						double delta = (pi_->unwrappedPos[pID*pi_->nDim + j] - pi_->crystalPos[pID*pi_->nDim + j]);
						energy += delta*delta;
					}
				}
				energy *= einsteinLambda/lamNorm;
			}
			break;
		case simControlStruct::einsteinEnum::periodic:
			{
				double ksp = 2.0 * PI / pi_->latticeSpacing;
				double normalization = getPeriodicRefNormalization();
				double lamNorm = getLamNorm();
				for (long pID = 0; pID < pi_->numParticles; pID++){

					double cosList = calcPeriodicCosList(pID, ksp, vector<double>(pi_->nDim,0.0));
					// energy += (1.0 - cosList)/(8.0 * PI * PI * meanRadius * meanRadius / (pi_->latticeSpacing * pi_->latticeSpacing));
					energy += cosList/normalization;
					// energy += 0.5*(1.0 - cosList)*(4 * meanRadius * meanRadius);
				}
				energy *= (einsteinLambda/lamNorm);
			}
			break;
		case simControlStruct::einsteinEnum::periodicUmbrella:
			{
				double w = 0;
				double ksp = 2.0 * PI / pi_->latticeSpacing;
				double normalization = getPeriodicRefNormalization();
				double lamNorm = getLamNorm();
				for (long pID = 0; pID < pi_->numParticles; pID++){
					double cosList = calcPeriodicCosList(pID, ksp, vector<double>(pi_->nDim,0.0));
					energy += (1.0 - cosList)/normalization;
					w += (1.0 + cosList)/normalization;
					// energy += (1.0 - cosList)/(1.0 + cosList)*einsteinLambda*pi_->numParticles * (meanRadius * meanRadius * PI * PI) / (pi_->latticeSpacing * pi_->latticeSpacing);
					// energy += 0.5*(1.0 - cosList)*(4 * meanRadius * meanRadius);
				}
				// w = 1.0;
				// energy = energy*exp(-w*einsteinLambda/lamNorm);
				energy = energy/w;
			}
			return energy;
		case simControlStruct::einsteinEnum::periodicUmbrellaExp:
			{

				double w = 0.0;
				double ksp = 2.0 * PI / pi_->latticeSpacing;
				double normalization = getPeriodicRefNormalization();
				double lamNorm = getLamNorm();
				for (long pID = 0; pID < pi_->numParticles; pID++){
					double cosList = calcPeriodicCosList(pID, ksp, vector<double>(pi_->nDim,0.0));
					w += cosList/normalization;
				}
				energy = w*exp(-w*einsteinLambda/lamNorm)/lamNorm;
			}
			return energy;
		default:
			break;
		}	
	return energy;
}

double MCPacking::getLamNorm(){
	double meanRadius = getMeanRadius();
	// if (pi_->nDim >= 9){
		// return (16.0 * meanRadius * meanRadius * meanRadius * meanRadius);
	// }else{
		return (4.0 * meanRadius * meanRadius);
	// }
	// return 1.0;
}

double MCPacking::getPeriodicRefNormalization(){
	double ksp = 2.0 * PI / pi_->latticeSpacing;
	// if (pi_->nDim >= 9){
	// 	return ksp * ksp * ksp * ksp / 2.0;
	// }else 
	if(pi_->nDim>=8){
		// return ksp * ksp;
		return ksp * ksp / 2.0;
	}else{
		return ksp * ksp / 2.0;
	}
	// return 2.0;
}


double MCPacking::getEinsteinWeightedConstant(){
	double energy = 0;
	double meanRadius = getMeanRadius();
	switch(pi_->simControl.einsteinType) {
		case simControlStruct::einsteinEnum::spring:
			{				
				energy = 1.0;
			}
			break;
		case simControlStruct::einsteinEnum::springUntethered:
			{				
				energy = 1.0;
			}
			break;
		case simControlStruct::einsteinEnum::periodic:
			{
				// double ksp = 2.0 * PI / pi_->latticeSpacing;
				// for (long pID = 0; pID < pi_->numParticles; pID++){
				// 	double cosList = calcPeriodicCosList(pID, ksp, vector<double>(pi_->nDim,0.0));
				// 	energy += (1.0/(1.0 + cosList)/(8.0 * PI * PI * meanRadius * meanRadius / (pi_->latticeSpacing * pi_->latticeSpacing)));
				// 	// energy += 0.5*(1.0 - cosList)*(4 * meanRadius * meanRadius);
				// }
				energy = 1.0;
			}
			break;
		case simControlStruct::einsteinEnum::periodicUmbrella:
			{
				double w = 0.0;
				double ksp = 2.0 * PI / pi_->latticeSpacing;
				double normalization = getPeriodicRefNormalization();
				double lamNorm = getLamNorm();
				for (long pID = 0; pID < pi_->numParticles; pID++){
					double cosList = calcPeriodicCosList(pID, ksp, vector<double>(pi_->nDim,0.0));
					w += (1.0 + cosList)/normalization;
				}
				energy = lamNorm/w;
			}
			return energy;
		case simControlStruct::einsteinEnum::periodicUmbrellaExp:
			{
				double w = 0.0;
				double ksp = 2.0 * PI / pi_->latticeSpacing;
				double normalization = getPeriodicRefNormalization();
				double lamNorm = getLamNorm();
				for (long pID = 0; pID < pi_->numParticles; pID++){
					double cosList = calcPeriodicCosList(pID, ksp, vector<double>(pi_->nDim,0.0));
					w += cosList/normalization;
				}
				energy = exp(-w*einsteinLambda/lamNorm);
			}
			return energy;
		default:
			break;
		}
	return einsteinLambda*energy;
}

bool MCPacking::singleMCEinstein(const long pID, const vector<double>& translationVector){
	bool makeMove = true;
	vector<double> translatedPos(pi_->nDim);
	/*
	We calculate the change in harmonic energy from the Einstein crystal as calculated in
	Section 10.3.2 of Understanding Molecular Simulation by Frenkel 
	*/

	double factor = 0.0;
	double meanRadius = getMeanRadius();
	double boltzmannFactor = 0.0;
	switch(pi_->simControl.einsteinType) {
		case simControlStruct::einsteinEnum::spring:
			{
				for (int j = 0; j < pi_->nDim; j++){
					factor += 2.0*(pi_->unwrappedPos[pID*pi_->nDim + j] - pi_->crystalPos[pID*pi_->nDim + j] - pi_->centerOfMassShift[j]) * translationVector[j] + double(((pi_->numParticles - 1)/double(pi_->numParticles))) * translationVector[j]*translationVector[j];
				}
				factor /= (4.0*meanRadius*meanRadius);
			}
			boltzmannFactor = exp(-einsteinLambda*factor);
			break;
		case simControlStruct::einsteinEnum::springUntethered:
			{
				// double factorNew = 0;
				// double factorCurrent = 0;
				double lamNorm = 1.0/(4.0 * meanRadius * meanRadius);
				for (int j = 0; j < pi_->nDim; j++){
					factor += (2.0*(pi_->unwrappedPos[pID*pi_->nDim + j] - pi_->crystalPos[pID*pi_->nDim + j]) + translationVector[j]) * translationVector[j];
					// factorCurrent += (pi_->unwrappedPos[pID*pi_->nDim + j] - pi_->crystalPos[pID*pi_->nDim + j]) * (pi_->unwrappedPos[pID*pi_->nDim + j] - pi_->crystalPos[pID*pi_->nDim + j]);
					// factorNew += (pi_->unwrappedPos[pID*pi_->nDim + j] - pi_->crystalPos[pID*pi_->nDim + j] + translationVector[j]) * (pi_->unwrappedPos[pID*pi_->nDim + j] - pi_->crystalPos[pID*pi_->nDim + j] + translationVector[j]);
				}
				// factor = factorNew - factorCurrent;
				boltzmannFactor = exp(-einsteinLambda*factor/lamNorm);
			}
			break;
		case simControlStruct::einsteinEnum::periodic:
			{
				double ksp = 2.0 * PI / pi_->latticeSpacing;
				double normalization = getPeriodicRefNormalization();
				double lamNorm = getLamNorm();
				double cosListCurrent = calcPeriodicCosList(pID, ksp, vector<double>(pi_->nDim,0.0));
				double cosListNew = calcPeriodicCosList(pID, ksp, translationVector);
				factor = (cosListNew - cosListCurrent)/normalization;	
				if (pi_->nDim==9){
					// Factor of 2 from occupancy
					// factor = factor / 2.0;
				}
				// factor = ((1.0 - cosListNew) - (1.0 - cosListCurrent))/(8.0 * PI * PI * meanRadius * meanRadius / (pi_->latticeSpacing * pi_->latticeSpacing));
				// factor = (0.5*(1.0 - cosListNew) - 0.5*(1.0 - cosListCurrent));
				boltzmannFactor = exp(-einsteinLambda*factor/lamNorm);
			}
			break;
			
		case simControlStruct::einsteinEnum::periodicUmbrella:
			{
				double ksp = 2.0 * PI / pi_->latticeSpacing;
				double normalization = getPeriodicRefNormalization();
				double lamNorm = getLamNorm();
				double wfactor = 0.0;
				double cosListCurrent = calcPeriodicCosList(pID, ksp, vector<double>(pi_->nDim,0.0));
				double cosListNew = calcPeriodicCosList(pID, ksp, translationVector);
				wfactor = (1.0 + cosListNew)/(1.0 + cosListCurrent);
				factor = (cosListCurrent - cosListNew)/normalization;
				boltzmannFactor = wfactor*exp(-factor*einsteinLambda/lamNorm);
			}
			break;
		case simControlStruct::einsteinEnum::periodicUmbrellaExp:
			{
				boltzmannFactor = 0.5;
			}	
			break;
		default:
			break;
		}
	
	// cout << "translation vector is [" << translationVector[0] << ", " << translationVector[1] << ", " << translationVector[2] << "]" << endl;
	// cout << "factor is " << boltzmannFactor << endl;
	
	// double boltzmannFactor = exp(-einsteinLambda*factor);
	double randomNumber = uniformDist(mt);
	if (boltzmannFactor > randomNumber){
		for (int i = 0; i < pi_->nDim; i++){
			translatedPos[i] = pi_->unwrappedPos[pID*pi_->nDim + i] + translationVector[i];
		}
		// if(pi_->simControl.geometryType == simControlStruct::geometryEnum::Dn){
		// 	vector<int32_t> mirror(pi_->nDim,0);
		// 	dnLatticePoint(translatedPos.begin(),mirror);
		// }
		switch(pi_->simControl.particleType) {
			case simControlStruct::particleEnum::cube:
				{
					if (!(pi_->simControl.monteCarloType==simControlStruct::monteCarloEnum::ecmc)){
						for(size_t i = 0; i < neighborList[pID].size(); i++){
							long nID = neighborList[pID][i];
							if (calcOverlap(translatedPos.begin(),pi_->unwrappedPos.begin() + nID * pi_->nDim, pID, nID) > 0.0){
								return false;
							}
						}
					}
					break;
				}

			default:
				{
					for(size_t i = 0; i < neighborList[pID].size(); i++){
						long nID = neighborList[pID][i];
						double radSum = getRadSum(pi_->radii[pID],pi_->radii[nID]);
						if (pbcDistanceSq(translatedPos.begin(),pi_->unwrappedPos.begin() + nID * pi_->nDim, pID, nID) < radSum*radSum) {
							return false;
						}
					}
					break;
				}
			}
	}

	makeMove = boltzmannFactor > randomNumber;
	if (makeMove) {
		pi_->translateParticle(pID,translationVector);
		shiftCenterOfMass(translationVector,pi_->numParticles);
	}
	return makeMove;
}

double MCPacking::calcPeriodicCosList(const long pID, const double ksp, const vector<double> translationVector){
	double cosList = 0.0;
	switch(pi_->simControl.particleType) {
		case simControlStruct::particleEnum::sphere:
			{
				cosList = 1.0;
				// if (pi_->nDim == 9){
				// 	double shift, sign;
				// 	if (pID < 19683){
				// 		shift = 0;
				// 		sign = 1;
				// 	}else{
				// 		shift = (1.0/6.0);
				// 		sign = -1;
				// 	}
				// 	for (int j = 0; j < pi_->nDim; j++){
				// 		cosList *= cos(ksp * (pi_->unwrappedPos[pID*pi_->nDim + j]  + shift + sign * d9Shift[j] + translationVector[j]));
				// 	}
				if(pi_->nDim >= 8){
					for (int j = 0; j < pi_->nDim; j++){
						cosList *= cos(ksp * (pi_->unwrappedPos[pID*pi_->nDim + j] + translationVector[j] - (1.0/6.0)));
					}
					double tempVariable = 1.0 - cosList;
					cosList = 1.0;
					for (int j = 0; j < pi_->nDim; j++){
						cosList *= cos(ksp * (pi_->unwrappedPos[pID*pi_->nDim + j] + translationVector[j]));
					}
					cosList = (1.0 - cosList) * tempVariable;
				}else{
					for (int j = 0; j < pi_->nDim; j++){
						cosList *= cos(ksp * (pi_->unwrappedPos[pID*pi_->nDim + j] + translationVector[j]));
					}
					cosList = 1.0 - cosList;
				}
				break;
			}
		case simControlStruct::particleEnum::cube:
			{
				double cosList = 0.0;
				for (int j = 0; j < pi_->nDim; j++){
					cosList += 1.0 + cos(ksp * (pi_->unwrappedPos[pID*pi_->nDim + j] + translationVector[j]));
				}
				return cosList;
			}
		}
	return cosList;
}

double MCPacking::calcGap(long pID, long nID){
	// vector<double> pos1(pi_->pos.begin() + pID * pi_->nDim, pi_->pos.begin() + (pID + 1) * pi_->nDim);
	// vector<double> pos2(pi_->pos.begin() + nID * pi_->nDim, pi_->pos.begin() + (nID + 1) * pi_->nDim);
	// double dist = pbcDistance(pi_->pos.begin() + pID * pi_->nDim,pi_->pos.begin() + nID * pi_->nDim, pID, nID);
	double radSum = getRadSum(pi_->radii[pID], pi_->radii[nID]);
	double thisOverlap = calcOverlap(pi_->pos.begin() + pID * pi_->nDim,pi_->pos.begin() + nID * pi_->nDim, pID, nID);
	switch(pi_->simControl.potentialType) {
		case simControlStruct::potentialEnum::log:
			return -thisOverlap;
			// return dist/radSum - 1.0;
		default:
			return -thisOverlap*radSum;
			// return dist - radSum;
		}
}

// double MCPacking::calcOverlap(long pID, long nID){
// 	return calcOverlap(pi_->pos.begin() + pID * pi_->nDim,pi_->pos.begin() + nID * pi_->nDim, pID, nID);
// 	// vector<double> pos1(pi_->pos.begin() + pID * pi_->nDim, pi_->pos.begin() + (pID + 1) * pi_->nDim);
// 	// vector<double> pos2(pi_->pos.begin() + nID * pi_->nDim, pi_->pos.begin() + (nID + 1) * pi_->nDim);
// 	// double dist = pbcDistance(pi_->pos.begin() + pID * pi_->nDim,pi_->pos.begin() + nID * pi_->nDim, pID, nID);
// 	// double radSum = pi_->radii[pID] + pi_->radii[nID];
// 	// return (1.0 - dist/radSum);
// }

double MCPacking::calcOverlap(vector<double>::iterator pos1, vector<double>::iterator pos2, long pID, long nID){
	double dist=0.0;
	switch(pi_->simControl.particleType) {
		case simControlStruct::particleEnum::sphere:{
			dist = pbcDistance(pos1,pos2,pID,nID);
			break;
		}
		case simControlStruct::particleEnum::cube:{
			vector<double> dispVector = pbcDisplacement(pos1, pos2, pID, nID);
			dist = abs(*max_element(dispVector.begin(),dispVector.end(),abs_compare));
			break;
		}
	}
	double radSum = getRadSum(pi_->radii[pID], pi_->radii[nID]);
	return (1.0 - dist/radSum);
}

bool MCPacking::anyOverlap(){
	for(long pID = 0; pID < pi_->numParticles-1; pID++){
		for(long nID = pID+1; nID < pi_->numParticles; nID++){
			double overlap = calcOverlap(pi_->unwrappedPos.begin() + pID * pi_->nDim,pi_->unwrappedPos.begin() + nID * pi_->nDim, pID, nID);
			if (overlap > 2.0e-15){
				// cout << overlap << endl;
				return true;
			}
		}
	}
	return false;
}

void MCPacking::MCStepECMC(int dir){
	uniform_int_distribution<int> volumeMoveDist(0,ceil(ecmcLength*pow(pi_->numParticles,1.0/pi_->nDim)));
	// cout << "here 00" << endl;
	long volumeMove = long(volumeMoveDist(mt));
	// cout << "here 0 " << endl;
	if (volumeMove==0 && (pi_->simControl.ensembleType == simControlStruct::ensembleEnum::npt)){
		// cout << "here 1" << endl;
		MCStepVolume();
	}else{
		uniform_int_distribution<int> particleDist(0,pi_->numParticles-1);
		long pID = long(particleDist(mt));
		uniform_int_distribution<int> parityDist(0,1);
		int parity = parityDist(mt);
		if (parity == 0){
			parity = -1.0;
		}else{
			parity = 1.0;
		}
		// cout << "here 2" << endl;
		double lengthTravelled = 0.0;
		double minRadius = *min_element(pi_->radii.begin(),pi_->radii.end());
		double numWraps = 0.0;
		double beforeX = pi_->pos[pID * pi_->nDim + dir];
		while (lengthTravelled < ecmcLength - 1e-15){
		// while (lengthTravelled < ecmcLength - 1e-15){
			// cout << "here 3" << endl;
			long collisionParticle = pID;
			double minDistToContact = 100.0;
			for(size_t i = 0; i < neighborList[pID].size(); i++){
				// cout << "here 4" << endl;
				long nID = neighborList[pID][i];
				vector<double> thisDispVec = pbcDisplacement(pi_->unwrappedPos.begin() + pID * pi_->nDim, pi_->unwrappedPos.begin() + nID * pi_->nDim, pID, nID);
				double radSum = getRadSum(pi_->radii[pID], pi_->radii[nID]);
				double thisDist = 0.0;
				for (int dim = 0; dim < pi_->nDim; dim++){
					if (dim != dir){
						if (abs(thisDispVec[dim]) > radSum){
							goto ENDDOUBLELOOP;
						}
					}else{
						if (parity * thisDispVec[dim] < 0.0){
							thisDist = abs(thisDispVec[dim]) - radSum;
						}else{
							thisDist = pi_->boxSize[dim] - abs(thisDispVec[dim]) - radSum;
						}
						if(thisDist > minDistToContact){
							goto ENDDOUBLELOOP;
						}
					}
				}			
				if (thisDist < ecmcLength - lengthTravelled){
					minDistToContact = thisDist;
					collisionParticle = nID;
				}
				ENDDOUBLELOOP:;
			}
			double particleDisplacement = min(ecmcLength - lengthTravelled,abs(minDistToContact));
			bool recalculateNeighbors = false;
			double neighborMoveDistance = pi_->lastNeighborCut * 2.0 * minRadius - getSumOfMaxDisplacements();
			if(particleDisplacement > neighborMoveDistance + 1e-15){
			// if(particleDisplacement > neighborMoveDistance){
				recalculateNeighbors = true;
				particleDisplacement = neighborMoveDistance;
			}
			// recalculateNeighbors = true;
			particleDisplacement = min(particleDisplacement,neighborMoveDistance);
			vector<double> translationVector = vector<double>(pi_->nDim,0.0);
			translationVector[dir] = double(parity)*particleDisplacement;
			double startBefore = pi_->pos[pID * pi_->nDim + dir];
			double colBefore = pi_->pos[collisionParticle * pi_->nDim + dir];
			pi_->translateParticle(pID,translationVector);
			// calcNeighborsCut(pi_->lastNeighborCut);
			// cout << "overlap? " << anyOverlap() << " " << getEnergy() << endl;
			double colAfter = pi_->pos[collisionParticle * pi_->nDim + dir];
			double startAfter = pi_->pos[pID * pi_->nDim + dir];
			numWraps += abs(floor((startBefore + double(parity)*particleDisplacement)/pi_->boxSize[dir]));
			if ((colAfter - startAfter) * double(parity) < 0.0){
				numWraps += 1.0;
			}
			lengthTravelled += particleDisplacement;
			// pID = collisionParticle;
			if (recalculateNeighbors){
				calcNeighborsCut(pi_->lastNeighborCut);
				collisionParticle = pID;
				resetDisplacement();
			}else{
				pID = collisionParticle;
			}
		}
		double totalDisp = 0.0;
		double afterX = pi_->pos[pID * pi_->nDim + dir];
		if (parity < 0.0){
			if (numWraps > 0){
				totalDisp = beforeX - (double(pi_->boxSize[dir]) - afterX) - (numWraps - 1.0)*double(pi_->boxSize[dir]);
				totalDisp = 0.0;
				totalDisp = beforeX + (numWraps*pi_->boxSize[dir] - afterX);
			}else{
				totalDisp = beforeX - afterX;
			}
		}else{
			if (numWraps > 0){
				totalDisp = afterX + numWraps * double(pi_->boxSize[dir]) - beforeX;
			}else{
				totalDisp = afterX - beforeX;
			}
		}
		ecmcPressure = abs(totalDisp)/ecmcLength;
		// cout << abs(totalDisp)/ecmcLength << endl;
	}
}

void MCPacking::MCStepMCMC(){
	// Generate the translation vector once and reuse it
	vector<double> translationVector(pi_->nDim);
	double minRadius = *min_element(pi_->radii.begin(),pi_->radii.end());
	bool accepted = false;
	long numAccepted = 0;
	long numSingleMCMoves = 0;
	long numMoves;
	if (pi_->simControl.ensembleType == simControlStruct::ensembleEnum::npt){
		// The final int in the intdist corresponds with a volume move, 
		// and second to last with a center of mass move			
		numMoves = pi_->numParticles+1;
	}else{
		// The final int in the intdist corresponds with a center of mass move			
		if(pi_->nDim == 9 && pi_->numParticles == 39366){
			numMoves = pi_->numParticles+9;
		}else{
			numMoves = pi_->numParticles;
		}
	}
	// The final int in the intdist corresponds with a center of mass move			
	uniform_int_distribution<int> intDist(0,numMoves);
	// Tell which type of Monte Carlo you'd like
	// TODO: Pass this in
	bool (MCPacking::*mcPtr)(const long,const vector<double>&);
	if (temperature!=0){
		mcPtr = &MCPacking::singleMCThermal;
	} else if (einsteinLambda!=0 && !(pi_->simControl.monteCarloType==simControlStruct::monteCarloEnum::ecmc)){
		mcPtr = &MCPacking::singleMCEinstein;
	} else{
		mcPtr = &MCPacking::singleMCHardSphere;
	}
	// Allow N+1 chances to move (N for particles, 1 for Center of Mass)
	for (long i = 0; i < pi_->numParticles + 1; i++){
		//TDOO: make this actually long dist
		long pID = long(intDist(mt));
		if (pID == pi_->numParticles){
			for (int j = 0; j < pi_->nDim; j++){
				if (!mcmcBlockedDirection[j]){
					translationVector[j] = 2.0 * maxDisplacementCM * minRadius * pi_->volume * (uniformDist(mt) - 0.5);
				}
			}
			// Only use single particle moves in the acceptance count
			accepted = centerOfMassMoveMC(translationVector);
		}else if(pID == pi_->numParticles+1 && pi_->simControl.ensembleType == simControlStruct::ensembleEnum::npt){
			accepted = MCStepVolume();
		}else if(pID > pi_->numParticles && pi_->nDim == 9 && pi_->numParticles == 39366){
			MCStepD9Relative(pID-(pi_->numParticles+1), 2.0 * maxDisplacementD9Relative * minRadius * pi_->volume * (uniformDist(mt) - 0.5));
		}else{
			for (int j = 0; j < pi_->nDim; j++){
				if (!mcmcBlockedDirection[j]){
					translationVector[j] = 2.0 * maxDisplacement * minRadius * pi_->volume * (uniformDist(mt) - 0.5);
				}
			}
			accepted = (this->*mcPtr)(pID,translationVector);
			numAccepted += long(accepted);
			numSingleMCMoves++;
		}
	}
	if(numSingleMCMoves == 0){
		acceptanceRatio = 0;
	}else{
		acceptanceRatio = double(numAccepted) / double(numSingleMCMoves);
	}
}

void MCPacking::MCStep(){
	switch(pi_->simControl.monteCarloType) {
		case simControlStruct::monteCarloEnum::ecmc:{
			uniform_int_distribution<int> dirDist(0,pi_->nDim-1);
			int dir = dirDist(mt);
			if (mcmcBlockedDirection[dir]){
				// cout << "this 1" << endl;
				// cout << int(floor(halfPowerPow(pi_->numRows,pi_->nDim-1))) << endl;
				// for(int i = 0; i < int(floor(halfPowerPow(pi_->numRows,pi_->nDim-1))); i++){
					MCStepECMC(dir);
				// }
			}else{
				// cout << "this 2" << endl;
				MCStepMCMC();
			}
			break;
		}
		default:{
			// cout << "this 3" << endl;
			MCStepMCMC();
			break;
		}
	}
}

void MCPacking::shiftCenterOfMass(const vector<double>& translationVector, long weight){
	for (int j = 0; j < pi_->nDim; j++){
		pi_->centerOfMassShift[j] += translationVector[j]/weight;
	}
}

void MCPacking::calcNeighborsCut(const double cutDistance) {
	pi_->lastNeighborCut = cutDistance;
	fill(neighborList.begin(), neighborList.end(), vector<long>(0));
	for (long pID = 0; pID < pi_->numParticles; pID++){
		for(long nID = pID+1; nID < pi_->numParticles; nID++){
			switch(pi_->simControl.particleType) {
				case simControlStruct::particleEnum::sphere:{
					// vector<double> pos2(pi_->pos.begin() + nID * pi_->nDim, pi_->pos.begin() + (nID + 1) * pi_->nDim);
					double thisDistSq = pbcDistanceSq(pi_->pos.begin() + pID * pi_->nDim,pi_->pos.begin() + nID * pi_->nDim, pID, nID);
					double radSum = getRadSum(pi_->radii[pID], pi_->radii[nID]);
					if (isNeighbor(thisDistSq,radSum,cutDistance)){
						neighborList[pID].push_back(nID);
						neighborList[nID].push_back(pID);
					}
					break;
				}
				case simControlStruct::particleEnum::cube:{
					vector<double> dispVector = pbcDisplacement(pi_->pos.begin() + pID * pi_->nDim, pi_->pos.begin() + nID * pi_->nDim, pID, nID);
					double dist = abs(*max_element(dispVector.begin(),dispVector.end(),abs_compare));
					double radSum = getRadSum(pi_->radii[pID], pi_->radii[nID]);
					if (dist < radSum*(1.0+cutDistance)){
						neighborList[pID].push_back(nID);
						neighborList[nID].push_back(pID);
					}
					break;
				}
			}
		}
	}
}

bool MCPacking::isNeighbor(double distSq, double radSum, double cutDistance){
	switch (pi_->simControl.potentialType) {
		case simControlStruct::potentialEnum::log: {
			double comparator = radSum*(1.0+(cutDistance*(1.0 + logZero) + logZero));
			return distSq < comparator*comparator;
		}
		case simControlStruct::potentialEnum::inversePower: {
			double comparator = radSum * (cutDistance + pi_->inversePowerCutoff);
			return distSq < comparator*comparator;
		}
		default: {
			double comparator = radSum*(1.0+cutDistance);
			return distSq < comparator*comparator;
		}
	}
}

bool MCPacking::isContact(double distSq, double radSum, double gapSize){
	// double overlap = 1.0 - dist/radSum;
	// switch(pi_->simControl.potentialType) {
	// 	case simControlStruct::potentialEnum::log:
	// 		return overlap > -logZero;
	// 	default:
	// 		return overlap > -gapSize;
	// }
	double comparator;
	switch(pi_->simControl.potentialType){
		case simControlStruct::potentialEnum::log:
			comparator = radSum*(1.0+logZero);
			break;
		case simControlStruct::potentialEnum::inversePower:
			// TODO: Come up with better definition of contact. Now contacts are less than cutoff
			comparator = radSum*pi_->inversePowerCutoff;
			break;
		default:
			comparator = radSum*(1.0+gapSize);
			break;
		}
	return distSq < comparator*comparator;
}

//Calculate and return the list of contacts
void MCPacking::calcContacts(double gap){
	fill(contactList.begin(), contactList.end(), vector<long>(0));
	for (long pID = 0; pID < pi_->numParticles; pID++){
		for(size_t i = 0; i < neighborList[pID].size(); i++){
			long nID = neighborList[pID][i];
			// vector<double> pos2(pi_->pos.begin() + nID * pi_->nDim, pi_->pos.begin() + (nID + 1) * pi_->nDim);
			double thisDistSq = pbcDistanceSq(pi_->pos.begin() + pID * pi_->nDim,pi_->pos.begin() + nID * pi_->nDim, pID, nID);
			double radSum = getRadSum(pi_->radii[pID], pi_->radii[nID]);
			if (isContact(thisDistSq,radSum,gap)){
				contactList[pID].push_back(nID);
			}
		}
	}
}

void MCPacking::calcEnergy(){
	// Also calculates the contact list, because it comes for free
	// #pragma omp parallel for 
	for(long pID = 0; pID < pi_->numParticles; pID++){
		energyVector[pID] = calcEnergyPerParticle(pID);
	}
}

void MCPacking::calcForceEnergy(){
	// Also calculates the contact list, because it comes for free
	// #pragma omp parallel for
	for(long pID = 0; pID < pi_->numParticles; pID++){
		energyVector[pID] = calcForceEnergyPerParticle(pID);
	}
	switch(pi_->simControl.constraintType) {
		case simControlStruct::constraintEnum::stressControl:
			for(long i = 0; i < pi_->numParticles * pi_->nDim; i++){
				pi_->netForce[i] += pi_->externalForce[i];
			}
			// transform(pi_->netForce.begin(), pi_->netForce.end(), pi_->externalForce.begin(), pi_->netForce.begin(), plus<double>());
			break;
		case simControlStruct::constraintEnum::strainControl:
			pi_->forceDotConstraint = -pi_->projectVector(pi_->netForce);
			break;
		default:
			break;
	}
}

double MCPacking::calcEnergyPerParticle(const long pID){
	double thisEnergy = 0.0;
	// vector<double> pos1(pi_->pos.begin() + pID * pi_->nDim, pi_->pos.begin() + (pID + 1) * pi_->nDim);
	// #pragma omp parallel for reduction(+: thisEnergy)
	for(size_t i = 0; i < neighborList[pID].size(); i++){
		long nID = neighborList[pID][i];
		// vector<double> pos2(pi_->pos.begin() + nID * pi_->nDim, pi_->pos.begin() + (nID + 1) * pi_->nDim);
		// double thisDist = pbcDistance(pi_->pos.begin() + pID * pi_->nDim,pi_->pos.begin() + nID * pi_->nDim, pID, nID);
		if (nID != pID){
			thisEnergy += calcEnergyPerInteraction(pi_->pos.begin() + pID * pi_->nDim,pi_->pos.begin() + nID * pi_->nDim, pID, nID);
		}
	}
	return thisEnergy;
}

double MCPacking::calcGradMultiple(const double norm, const double overlap, const double radSum) {
	switch (pi_->simControl.potentialType) {
		case simControlStruct::potentialEnum::contactPower: {
			if (pi_->potentialPower == 2) {
				return EPSILON * (overlap > 0) * overlap / (radSum * norm);
			} else {
				if (overlap > 0){
					return EPSILON * halfPowerPow(overlap, pi_->potentialPower - 1.0) / (radSum * norm);
				} else{
					return 0.0;
				}
			}
		}
		case simControlStruct::potentialEnum::log: {
			if (overlap >= 0) {
				return NAN;
			}	else {
				if (overlap > -logZero) {
				// If we're within d_logZero distance then the force is inverse to the overlap shifted to hit zero at d_logZero
					return (overlap < 0)*(-1.0/(overlap*radSum)) - (overlap < 0)/(logZero*radSum);
				} else {
				// If we're beyond d_logZero then the force is just zero
					return 0;
				}
			}
		}
		case simControlStruct::potentialEnum::inversePower: {
			// distRatio is distance / radSum
			double distRatio = 1.0 - overlap;
			return EPSILON * (distRatio<pi_->inversePowerCutoff) * (double(pi_->inversePower)*pow(1.0/distRatio,pi_->inversePower+1) - 2.0*pi_->inversePowerC1*distRatio - 4.0*pi_->inversePowerC2*pow(distRatio,3)) / (radSum * norm);
		}
		default:
			return NAN;
	}
}

void MCPacking::calcDeltaNormSq(vector<double>::iterator x, vector<double>::iterator y, vector<double>::iterator delta, double& normSq) {
	switch (pi_->simControl.geometryType) {
		case simControlStruct::geometryEnum::leesEdwards:
			for (int dim = 0; dim < pi_->nDim; dim++) {
				if (dim == 1) {
					delta[dim] = pbcDistanceLE(x[0], x[1], y[0], y[1]);
				}
				else {
					delta[dim] = pbcDistance(x[dim], y[dim]);
				}
				normSq += delta[dim] * delta[dim];
			}
			break;
		default:
			for (int dim = 0; dim < pi_->nDim; dim++) {
				delta[dim] = pbcDistance(x[dim], y[dim], pi_->boxSize[dim]);
				normSq += delta[dim] * delta[dim];
			}
			break;
	}
}

double MCPacking::pbcDistanceLE(const double x1, const double x2, const double y1, const double y2) {
	double delta1 = (x1 - y1);
	// double rounded = rint(delta1); //need to store for lees-edwards BC
	double rounded = fastRound(delta1); //need to store for lees-edwards BC
	delta1 -= rounded;
	double delta2 = (x2 - y2);
	// delta2 = delta2  - rounded*pi_->LEshift - rint(delta2 - rounded*pi_->LEshift);
	delta2 = delta2  - rounded*pi_->LEshift - fastRound(delta2 - rounded*pi_->LEshift);
	return delta2;
}

vector<double> MCPacking::getPerParticleStressTensor(){
	vector<double> stressTensor = vector<double>(pi_->numParticles*pi_->nDim*pi_->nDim,0.0);
	for(long pID = 0; pID < pi_->numParticles; pID++){
		for(size_t j = 0; j < neighborList[pID].size(); j++){
			long nID = neighborList[pID][j];	
			vector<double> delta = vector<double>(pi_->nDim,0.0);
			auto normSq = 0.0;
			// TODO: Fix for cubes
			calcDeltaNormSq(pi_->pos.begin() + pID * pi_->nDim,pi_->pos.begin() + nID * pi_->nDim, delta.begin(), normSq);
			double norm = sqrt(normSq);
			double radSum = getRadSum(pi_->radii[pID], pi_->radii[nID]);
			double overlap = 1.0 - norm/radSum;
			double gradMultiple = calcGradMultiple(norm, overlap, radSum);
			for(int alpha = 0; alpha < pi_->nDim; alpha++){
				for(int beta = 0; beta < pi_->nDim; beta++){
					//Since force = delta*gradMultiple we can simplify by using
					//delta[alpha]*force[beta] = delta[alpha]*delta[beta]*gradMultiple
					//We divide by 2 because we are double counting every contact
					stressTensor[pID*pi_->nDim*pi_->nDim + alpha*pi_->nDim + beta] += -pi_->phi*delta[alpha]*delta[beta]*gradMultiple/2.0;  
				}
			}
		}
	}
	return stressTensor;
}

void MCPacking::setLogZero(const double logZero_i){
	logZero = logZero_i;
}

void MCPacking::setNumRows(const int numRows_i){
	pi_->numRows = numRows_i;
}

int MCPacking::getNumRows(){
	return pi_->numRows;
}

void MCPacking::setNonAdditivity(const double nonAdditivity_i){
	pi_->nonAdditivity = nonAdditivity_i;
}

double MCPacking::getNonAdditivity(){
	return pi_->nonAdditivity;
}

void MCPacking::setECMCLength(const double ecmcLength_i){
	ecmcLength = ecmcLength_i;
}

double MCPacking::getECMCLength(){
	return ecmcLength;
}

double MCPacking::getECMCPressure(){
	return ecmcPressure;
}

double MCPacking::calcForceEnergyPerParticle(const long pID){
	double thisEnergy = 0.0;
	for(int dimIndex = 0; dimIndex < pi_->nDim; dimIndex++){
		pi_->netForce[pID*pi_->nDim + dimIndex] = 0.0;
	}
	double forceMag = 0.0;
	// vector<double> pos1(pi_->pos.begin() + pID * pi_->nDim, pi_->pos.begin() + (pID + 1) * pi_->nDim);
	// #pragma omp parallel for reduction(+: thisEnergy, forceMag)
	for(size_t i = 0; i < neighborList[pID].size(); i++){
		long nID = neighborList[pID][i];		
		// TODO: combine these functions
		thisEnergy += calcEnergyPerInteraction(pi_->pos.begin() + pID * pi_->nDim,pi_->pos.begin() + nID * pi_->nDim, pID, nID);
		forceMag = calcForcePerInteraction(pi_->pos.begin() + pID * pi_->nDim,pi_->pos.begin() + nID * pi_->nDim, pID, nID);
		vector<double> nVector = calcNVector(pID,nID);
		for (int dimIndex = 0; dimIndex < pi_->nDim; dimIndex++){
			pi_->netForce[pID*pi_->nDim + dimIndex] += -forceMag * nVector[dimIndex];
		}
	}
	return thisEnergy;
}

tuple< vector<long>, vector<long> > MCPacking::getContacts(double gap) {
	//Calculate the set of contacts
	calcContacts(gap);
	vector<long> indices;
	vector<long> indptr;
	pi_->doubleVectorToCSR(contactList, indices, indptr);
	return make_tuple(indices, indptr);
}

tuple< vector<long>, vector<long>, vector<double> > MCPacking::getContactVectors(double gap) {
	vector<long> indices1;
	vector<long> indices2;
	vector< vector<double> > contactVectorVector(pi_->numParticles);
	vector<double> contactVectorsData;
	//Calculate the set of contacts
	calcContacts(gap);
	// #pragma omp parallel for
	for (long pID = 0; pID < pi_->numParticles; pID++){
		for(size_t j = 0; j < contactList[pID].size(); j++){
			long nID = contactList[pID][j];
			vector<double> thisContactVector = pbcDisplacement(pi_->pos.begin() + pID * pi_->nDim,pi_->pos.begin() + nID * pi_->nDim, pID, nID);
			for(int k = 0; k < pi_->nDim; k++){
				contactVectorVector[pID].push_back(thisContactVector[k]);
			}
		}
	}
	pi_->doubleVectorToCOO(contactList, contactVectorVector, indices1, indices2, contactVectorsData, 1, pi_->nDim);
	return make_tuple(indices1, indices2, contactVectorsData);
}

void MCPacking::setTargetPressure(double const targetPressure_i){
	pi_->targetPressure = targetPressure_i;
}

double MCPacking::getTargetPressure(){
	return pi_->targetPressure;
}

tuple< vector<long>, vector<long>, vector<double> > MCPacking::getSmallGaps(double maxGap){
	vector< vector<double> > gapList(pi_->numParticles);
	vector< vector<long> > gapIndices(pi_->numParticles);
	vector<double> gapsData;
	vector<long> indices1;
	vector<long> indices2;
	pi_->calcSmallGaps(gapList, gapIndices, maxGap);
	pi_->doubleVectorToCOO(gapIndices, gapList, indices1, indices2, gapsData, 1, 1);
	return make_tuple(indices1, indices2, gapsData);
}

tuple< vector<long>, vector<long>, vector<double> > MCPacking::getSmallOverlaps(double minOverlapData){
	vector< vector<double> > overlapList(pi_->numParticles);
	vector< vector<long> > overlapIndices(pi_->numParticles);
	vector<long> indices1;
	vector<long> indices2;
	vector<double> overlapsData;
	pi_->calcSmallOverlaps(overlapList, overlapIndices, minOverlapData);
	pi_->doubleVectorToCOO(overlapIndices, overlapList, indices1, indices2, overlapsData, 1, 1);
	return make_tuple(indices1, indices2, overlapsData);
}

double MCPacking::calcEnergyPerInteraction(const vector<double>::iterator pos1, const vector<double>::iterator pos2, const long pID, const long nID){
	// double thisOverlap = 1.0 - (distance/(radius1 + radius2));
	double thisOverlap = calcOverlap(pos1,pos2,pID,nID);
	switch(pi_->simControl.potentialType) {
		case simControlStruct::potentialEnum::contactPower:
			if (pi_->potentialPower == 2.0) {
				return EPSILON * thisOverlap * thisOverlap * (thisOverlap > 0.0)/2.0;
			} else {
				if (thisOverlap > 0){
					return EPSILON*halfPowerPow(thisOverlap,pi_->potentialPower)/pi_->potentialPower;
				} else{
					return 0.0;
				}
			}
		case simControlStruct::potentialEnum::log:
			if (thisOverlap >= 0.0) {
				return std::numeric_limits<double>::infinity();
			} else {
				if (thisOverlap > -logZero){
					return (thisOverlap<0.0)*((1.0 - thisOverlap)/logZero - log(-thisOverlap));
				} else {
					return 0.0;
				}
			}
		case simControlStruct::potentialEnum::inversePower:
			{
				double distRatio =  1.0 - thisOverlap;
				return EPSILON * (distRatio<pi_->inversePowerCutoff) * (halfPowerPow(1.0/distRatio,pi_->inversePower) + pi_->inversePowerC0 + pi_->inversePowerC1*halfPowerPow(distRatio,2) + pi_->inversePowerC2*halfPowerPow(distRatio,4));
			}
		default:
			return NAN;
	}
	
}

double MCPacking::halfPowerPow(const double base, const double powerNum){
	int powNumInt = fastRound(2.0*powerNum);
	switch (powNumInt) {
		case 1:
			return sqrt(base);
		case 2:
			return base;
		case 3:
			return base * sqrt(base);
		case 4:
			return base * base;
		case 5:
			return base * base * sqrt(base);
		case 6:
			return base * base * base;
		case 7:
			return base * base * base * sqrt(base);
		case 8:
			return base * base * base * base;
		case 9:
			return base * base * base * base * sqrt(base);
		case 10:
			return base * base * base * base * base;
		default:
			return pow(base, powerNum);
	}
}

double MCPacking::calcForcePerInteraction(const vector<double>::iterator pos1, const vector<double>::iterator pos2, const long pID, const long nID){
	// TODO: Simplify with calcGradMultiple
	double radSum = getRadSum(pi_->radii[pID], pi_->radii[nID]);
	double thisOverlap = calcOverlap(pos1,pos2,pID,nID);
	switch(pi_->simControl.potentialType) {
		case simControlStruct::potentialEnum::contactPower:
			if (pi_->potentialPower == 2.0) {
				return EPSILON*(thisOverlap/radSum)  * (thisOverlap > 0.0);
			} else {
				if (thisOverlap > 0.0){
					return EPSILON*halfPowerPow(thisOverlap,pi_->potentialPower-1.0)/radSum;
				} else{
					return 0.0;
				}
			}
		case simControlStruct::potentialEnum::log:
			if (thisOverlap >= 0.0) {
				return std::numeric_limits<double>::infinity();
			} else {
				if(thisOverlap > -logZero) {
					return -(thisOverlap < 0.0)*(1.0/logZero + 1.0/thisOverlap)*(1.0 - thisOverlap);
				} else {
					return 0.0;
				}
			}
		case simControlStruct::potentialEnum::inversePower:
			{
				// distRatio is distance / radSum
				double distRatio = 1.0 - thisOverlap;
				return EPSILON * (distRatio<pi_->inversePowerCutoff) * (double(pi_->inversePower)*halfPowerPow(1.0/distRatio,pi_->inversePower+1) - 2.0*pi_->inversePowerC1*distRatio - 4.0*pi_->inversePowerC2*halfPowerPow(distRatio,3)) / (radSum);
			}
		default:
			return NAN;
	}

}

double MCPacking::getEnergy(){
	// We're abandoning a smart sum (j < i) for per particle functionality. 
	// This double counts, so we must multiply by 0.5
	double energySum = 0.0;
	// #pragma omp parallel for reduction(+: energySum)
	for (long i = 0; i < pi_->numParticles; i++){
		energySum += energyVector[i];
	}
	return 0.5*energySum;
}

vector<double> MCPacking::getEnergyVector(){
	return energyVector;
}

double MCPacking::getLogZero(){
	return logZero;
}

void MCPacking::load(const string basedir){
	// Assumes that the position and radii files are basedir/positions.dat 
	// and basedir/radii.dat respectively
	
	// For now, we will assume that the user is smart and has input the correct
	// dimensions and number of particles for the system that is being loaded.
	// TODO: 	Fix this to set dimension and number of particles by reading a
	// 			scalars file that is in the same format as given in pcp.
	ifstream posFile;
	string line;
	posFile.open(basedir + "/positions.dat");
	if(posFile.is_open()){
		size_t numLines = 0;
		string token;
		while (getline(posFile,line)) {
			istringstream iss(line);
			while (getline(iss,token,'\t')) {
				pi_->pos[numLines] = stod(token);
				pi_->unwrappedPos[numLines] = pi_->pos[numLines];
				numLines++;
			}
		}
	} else {
		cout << "Could not find position file at " + basedir + "/positions.dat" << endl;
	}
	posFile.close();
	ifstream radFile;
	radFile.open(basedir + "/radii.dat");
	if(radFile.is_open()){
		size_t numLines = 0;
		while (getline(radFile,line)) {
			pi_->radii[numLines] = stod(line);
			numLines++;
		}
	} else {
		cout << "Could not find radius file at " + basedir + "/radii.dat" << endl;
	}
	radFile.close();
}

void MCPacking::setMCOverlap(double mcOverlap_i){
	pi_->overlapParameter = mcOverlap_i;
}

double MCPacking::getMCOverlap(){
	return pi_->overlapParameter;
}

double MCPacking::calcMCOverlap(vector<double>& initialPos){
	double accum = 0.0;

	// for (long i = 0; i < pi_->numParticles; i++){
	// 	for (int dimIndex = 0; dimIndex < pi_->nDim; dimIndex++){
	// 		initialPos[i * pi_->nDim + dimIndex] = initialPos[i * pi_->nDim + dimIndex] + pi_->centerOfMassShift[dimIndex];
	// 	}
	// }

	// #pragma omp parallel for reduction(+: accum)
	for (long pID = 0; pID < pi_->numParticles; pID++){
		// vector<double> thisPos(pi_->pos.begin() + i * pi_->nDim, pi_->pos.begin() + (i + 1) * pi_->nDim);
		// vector<double> iPos(initialPos.begin() + i * pi_->nDim, initialPos.begin() + (i + 1) * pi_->nDim);
		
		// Here we don't care about periodic distance, we want total distance travelled
		double dist = 0.0;
		for (int dim = 0; dim < pi_->nDim; dim++){
			double delta = pi_->unwrappedPos[pID * pi_->nDim + dim] - initialPos[pID * pi_->nDim + dim] - pi_->centerOfMassShift[dim];
			dist += delta * delta;
		}
		// dist += pbcDistanceSq(pi_->unwrappedPos.begin() + i * pi_->nDim, initialPos.begin() + i * pi_->nDim,-1,-1);
		// dist += (pi_->unwrappedPos[i * pi_->nDim + dimIndex] - initialPos[i * pi_->nDim + dimIndex] - pi_->centerOfMassShift[dimIndex]) * (pi_->unwrappedPos[i * pi_->nDim + dimIndex] - initialPos[i * pi_->nDim + dimIndex] - pi_->centerOfMassShift[dimIndex]);
		dist = sqrt(dist);
		accum += (2.0 * pi_->overlapParameter * pi_->radii[pID] - dist) > 0.0;
	}

	return accum/double(pi_->numParticles);
}

double MCPacking::getMaxUnbalancedForce(){
	double maxForce = 0.0;
	// #pragma omp parallel for reduction(max: maxForce)
	for(long pID = 0; pID < pi_->numParticles; pID++){
		double accum = 0.0;
		for(int dimIndex = 0; dimIndex<pi_->nDim; dimIndex++){
			accum += (pi_->netForce[pID*pi_->nDim + dimIndex] * pi_->netForce[pID*pi_->nDim + dimIndex]);
		}
		maxForce = max(maxForce,accum);
	}
	return sqrt(maxForce);
}

double MCPacking::getRadSum(double rad1, double rad2){
	switch(pi_->simControl.distanceType) {
		case simControlStruct::distanceEnum::nonAdditive:
			return (rad1 + rad2) * (1.0-pi_->nonAdditivity*abs(rad1 - rad2));
		default:
			return rad1 + rad2;
		}
}

double MCPacking::getSumOfMaxDisplacements(){
	vector<double> distArray(pi_->numParticles);
	std::fill(distArray.begin(), distArray.end(), 0.0);
	for (long i = 0; i < pi_->numParticles; i++){
		// vector<double> unPos(pi_->unwrappedPos.begin() + i * pi_->nDim, pi_->unwrappedPos.begin() + (i + 1) * pi_->nDim);
		// vector<double> lPos(pi_->lastPos.begin() + i * pi_->nDim, pi_->lastPos.begin() + (i + 1) * pi_->nDim);
		// distArray[i] = pbcDistanceSq(pi_->unwrappedPos.begin() + i * pi_->nDim,pi_->lastPos.begin() + i * pi_->nDim);
		distArray[i] = pbcDistanceSq(pi_->unwrappedPos.begin() + i * pi_->nDim, pi_->lastPos.begin() + i * pi_->nDim,-1,-1);
		// for (int dimIndex = 0; dimIndex < pi_->nDim; dimIndex++){
		// 	distArray[i] += (pi_->unwrappedPos[i * pi_->nDim + dimIndex] - pi_->lastPos[i * pi_->nDim + dimIndex]) * (pi_->unwrappedPos[i * pi_->nDim + dimIndex] - pi_->lastPos[i * pi_->nDim + dimIndex]);
		// }
	}
	sort(distArray.begin(),distArray.end());
	return sqrt(distArray[pi_->numParticles - 1]) + sqrt(distArray[pi_->numParticles - 2]);

	// if (pi_->distTravelledSq[0] > pi_->distTravelledSq[1]){
	// 	largest = pi_->distTravelledSq[0];
	// 	second = pi_->distTravelledSq[1];
	// } else {
	// 	largest = pi_->distTravelledSq[1];
	// 	second = pi_->distTravelledSq[0];
	// }
	// for(long i = 2; i < pi_->numParticles; i++){
	// 	if (pi_->distTravelledSq[i] > largest){
	// 		second = largest;
	// 		largest = pi_->distTravelledSq[i];
	// 	} else if (pi_->distTravelledSq[i] > second && pi_->distTravelledSq[i] != largest){
	// 		second = pi_->distTravelledSq[i];
	// 	}
	// }
	// return sqrt(largest) + sqrt(second);
}

double MCPacking::getMaxDisplacement(){
	vector<double> distArray(pi_->numParticles);
	std::fill(distArray.begin(), distArray.end(), 0.0);
	for (long i = 0; i < pi_->numParticles; i++){
		// vector<double> unPos(pi_->unwrappedPos.begin() + i * pi_->nDim, pi_->unwrappedPos.begin() + (i + 1) * pi_->nDim);
		// vector<double> lPos(pi_->lastPos.begin() + i * pi_->nDim, pi_->lastPos.begin() + (i + 1) * pi_->nDim);
		distArray[i] = pbcDistanceSq(pi_->unwrappedPos.begin() + i * pi_->nDim,pi_->lastPos.begin() + i * pi_->nDim,-1,-1);
		// for (int dimIndex = 0; dimIndex < pi_->nDim; dimIndex++){
		// 	distArray[i] += (pi_->unwrappedPos[i * pi_->nDim + dimIndex] - pi_->lastPos[i * pi_->nDim + dimIndex]) * (pi_->unwrappedPos[i * pi_->nDim + dimIndex] - pi_->lastPos[i * pi_->nDim + dimIndex]);
		// }
	}
	double maxDisp = *max_element(distArray.begin(),distArray.end());
	return sqrt(maxDisp);
}

void MCPacking::resetDisplacement(){
	for (long i = 0; i < pi_->nDim * pi_->numParticles; i++){
		pi_->lastPos[i] = pi_->unwrappedPos[i];
	}
}

double MCPacking::getInversePowerCutoff(){
	return pi_->inversePowerCutoff;
}

double MCPacking::hyperBallCoefficient(int dim){
	return pow(3.141592653589793238463,dim/2.0) / tgamma(dim/2.0 + 1);
}

vector<double> MCPacking::calcGRBinVolume(vector<double> binEdges,long numVolumePoints){
	int numSlices = int(binEdges.size()-1);
	vector<double> volumeSlice(numSlices);
	vector<double> binEdgeSq(numSlices+1);
	for(int i = 0; i < numSlices; i++){
		volumeSlice[i] = (pow(binEdges[i+1],pi_->nDim) - pow(binEdges[i],pi_->nDim))*hyperBallCoefficient(pi_->nDim);
		binEdgeSq[i] = binEdges[i]*binEdges[i];
	}
	binEdgeSq[numSlices] = binEdges[numSlices]*binEdges[numSlices];
	double lowestBinSq = binEdges[0]*binEdges[0];
	double highestBinSq = binEdges.back()*binEdges.back();
	double minBoxLength = *min_element(pi_->boxSize.begin(),pi_->boxSize.end());
	if (binEdges.back() > minBoxLength/2.0){
		vector<long> approxVolumeSlice(numSlices,0.0);
		double disp;
		for (long pID = 0; pID < numVolumePoints; pID++){
			double accum = 0.0;
			switch(pi_->simControl.geometryType) {
				case simControlStruct::geometryEnum::E8:{
						vector<int32_t> mirror(pi_->nDim,0);
						vector<double> point(pi_->nDim,0);
						vector<double> zeroVec(pi_->nDim,0);
						for (int dimIndex = 0; dimIndex < pi_->nDim; dimIndex++){
							point[dimIndex] = 2.0*(uniformDist(mt) - 0.5);
						}
						accum = pbcDistanceSq(point.begin(),zeroVec.begin(),-1,-1);
						break;
					}
				case simControlStruct::geometryEnum::Dn:{
						vector<int32_t> mirror(pi_->nDim,0);
						vector<double> point(pi_->nDim,0);
						vector<double> zeroVec(pi_->nDim,0);
						for (int dimIndex = 0; dimIndex < pi_->nDim; dimIndex++){
							point[dimIndex] = 2.0*(uniformDist(mt) - 0.5);
						}
						accum = pbcDistanceSq(point.begin(),zeroVec.begin(),-1,-1);
						break;
					}
				default:
					for (int dimIndex = 0; dimIndex < pi_->nDim; dimIndex++){
						disp = (uniformDist(mt) - 0.5)*pi_->boxSize[dimIndex];
						accum += disp*disp;
					}
					break;
			}
			//We now have the point distance, so find which bin it lives in
			if (accum < highestBinSq && accum > lowestBinSq){
				approxVolumeSlice[lower_bound(binEdgeSq.begin(),binEdgeSq.end()-1,accum) - binEdgeSq.begin()-1]++;
			}
		}
		// If our shell is larger than the min box length, use approx
		double upperCutoff;
		switch(pi_->simControl.geometryType) {
			case simControlStruct::geometryEnum::E8:
				upperCutoff = 0.25;
				break;
			case simControlStruct::geometryEnum::Dn:
				upperCutoff = 0.5;
				break;
			default:
				upperCutoff = minBoxLength*minBoxLength/4.0;
				break;
		}
		for (int i = 0; i < numSlices; i++){
			if(binEdges[i+1] > upperCutoff){
				volumeSlice[i] = pi_->volume * double(approxVolumeSlice[i]) / double(numVolumePoints);
			}
		}
	}
	return volumeSlice;
}

vector<double> MCPacking::calcGofR(const vector<double> binEdges, const long numVolumePoints){
	int numSteps = int(binEdges.size()-1);
	vector<double> binEdgeSq(numSteps+1);
	vector<double> g(numSteps,0.0);
	vector<long> gCount(numSteps,0.0);
	for(int i = 0; i < numSteps+1; i++){
		binEdgeSq[i] = binEdges[i]*binEdges[i];
	}
	vector<double> volumeSlice = calcGRBinVolume(binEdges,numVolumePoints);
	for (long pID = 0; pID < pi_->numParticles; pID++){
		// vector<double> pos1(pi_->pos.begin() + pID * pi_->nDim, pi_->pos.begin() + (pID + 1) * pi_->nDim);
		for(size_t j = 0; j < neighborList[pID].size(); j++){
			long nID = neighborList[pID][j];
			if (nID > pID){
				// vector<double> pos2(pi_->pos.begin() + nID * pi_->nDim, pi_->pos.begin() + (nID + 1) * pi_->nDim);
				double distanceSq = pbcDistanceSq(pi_->pos.begin() + pID * pi_->nDim,pi_->pos.begin() + nID * pi_->nDim, pID, nID);
				if (distanceSq < binEdgeSq.back() && distanceSq > binEdgeSq[0]){
					gCount[lower_bound(binEdgeSq.begin(),binEdgeSq.end()-1,distanceSq) - binEdgeSq.begin()-1]++;
				}
			}
		} 
	}
	for (int i=0; i < numSteps; i++){
		g[i] = double(gCount[i]) * pi_->volume / (double(pi_->numParticles * pi_->numParticles) * volumeSlice[i] * 0.5);
		// g[i] = double(gCount[i]);
		// cout << volumeSlice[i] << endl;
	}
	return g;
}

void MCPacking::save(const string basedir){
	// Saves positions and radii as basedir/positions.dat and basedir/radii.dat

	// TODO: 	Fix this to set dimension and number of particles by reading a
	// 			scalars file that is in the same format as given in pcp.

	// This option works with g++ 8, but for now I'm disabling it and assuming
	// that the directory tree is made by the python interface
	// if (std::filesystem::create_directories(basedir)) {
	// 	cout << "Had to create directory tree " << basedir << endl;
	// }
	ofstream posFile;
	posFile.open(basedir + "/positions.dat");
	for (size_t i = 0; i < pi_->pos.size(); i++) {
		posFile << std::fixed << std::setprecision(18) << pi_->pos[i];
		if (i % pi_->nDim==size_t(pi_->nDim-1)){
			posFile << "\n";
		} else {
			posFile << "\t";
		}
	}
	posFile.close();
	ofstream radFile;
	radFile.open(basedir + "/radii.dat");
	for(size_t i = 0; i < pi_->radii.size(); i++){
		radFile << std::fixed << std::setprecision(18) << pi_->radii[i] << "\n";
	}
	radFile.close();
	saveScalars(basedir);
}

void MCPacking::saveScalars(const string basedir){
	ofstream boxFile;
	boxFile.open(basedir + "/boxSize.dat");
	for(int i = 0; i < pi_->nDim; i++){
		boxFile << pi_->boxSize[i] << "\n";
	}
	boxFile.close();
	ofstream scalarFile;
	scalarFile.open(basedir + "/scalars.dat");
	scalarFile << "nDim\t" << pi_->nDim << endl;
	scalarFile << "numParticles\t" << pi_->numParticles << endl;
	scalarFile << "phi\t" << std::fixed << std::setprecision(18) << pi_->phi << endl;
	scalarFile << "potentialPower\t" << pi_->potentialPower << endl;
	scalarFile << "isFinished\t" << 1 << endl;
	scalarFile.close();
}

